""" Classes to do plots of kx and ky spectra"""

import matplotlib.pyplot as plt
import numpy as np
import warnings
from multiprocessing.dummy import Pool

from .baseplot import Plotting
from ..data.slices import MomFieldSlice
from ..utils.comm import DiagSpace

import ..utils.averages as averages
import ..utils.geom


class AmplitudeSpectra(object):
    """ kx and ky spectra of 3d mom and field data

     Collect the squared absolute field and mom data as kx and ky spectra
     """

    def __init__(self, common, species, rundatafiles, moms=None):
        if not common.y_local:
            raise NotImplementedError("Only implemented for local simulations yet")
        self.cm = common
        self.species = species
        self.diagspace_kx = DiagSpace(common.spatialgrid, x_fourier=True, y_fourier=True, yavg=True,
                                      zavg=True)
        self.diagspace_ky = DiagSpace(common.spatialgrid, x_fourier=True, y_fourier=True, xavg=True,
                                      zavg=True)
        self.momamps_kx = {"phi": None, "dens": None, "tpar": None, "tperp": None, "qpar": None,
                           "qperp": None, "upar": None}
        self.momamps_ky = {"phi": None, "dens": None, "tpar": None, "tperp": None, "qpar": None,
                           "qperp": None, "upar": None}
        if common.electromagnetic:
            self.momamps_kx.update({"apar": None})
            self.momamps_ky.update({"apar": None})
            if common.bpar:
                self.momamps_kx.update({"bpar": None})
                self.momamps_ky.update({"bpar": None})
        if moms:
            self.momamps_kx = {mom: self.momamps_kx[mom] for mom in moms if mom in self.momamps_kx}
            self.momamps_ky = {mom: self.momamps_ky[mom] for mom in moms if mom in self.momamps_ky}

        with Pool(len(self.momamps_kx)) as p:
            momamps_kx_temp = p.map(
                lambda mom: MomFieldSlice(common, mom, species, self.diagspace_kx, rundatafiles),
                self.momamps_kx)
            momamps_ky_temp = p.map(
                lambda mom: MomFieldSlice(common, mom, species, self.diagspace_ky, rundatafiles),
                self.momamps_ky)
        for imom, mom in enumerate(self.momamps_kx):
            self.momamps_kx[mom] = momamps_kx_temp[imom]
            self.momamps_kx[mom].generate_timeseries()
            self.momamps_kx[mom].generate_timeaverage()
            self.momamps_ky[mom] = momamps_ky_temp[imom]
            self.momamps_ky[mom].generate_timeseries()
            self.momamps_ky[mom].generate_timeaverage()

class PerpSpectra(object):
    """ 1d perpendicular spectra of 3d mom and field data

     Collect the squared absolute field and mom data as kperp spectra
     """

    def __init__(self, common, species, rundatafiles):
        if not common.x_local or not common.y_local:
            raise NotImplementedError("Only implemented for local simulations yet")
        self.cm = common
        self.species = species
        self.diagspace_kxky = DiagSpace(common.spatialgrid, x_fourier=True, y_fourier=True, zavg=True)
        self.momamps_kperp = {"phi": None}#, "dens": None, "tpar": None, "tperp": None, "qpar": None,
#                           "qperp": None, "upar": None}
        if common.electromagnetic:
            self.momamps_kperp.update({"apar": None})
            if common.bpar:
                self.momamps_kperp.update({"bpar": None})

        with Pool() as p:
            momamps_kperp_temp = p.map(
                lambda mom: MomFieldSlice(common, mom, species, self.diagspace_kxky, rundatafiles),
                self.momamps_kperp)
        for imom, mom in enumerate(self.momamps_kperp):
            self.momamps_kperp[mom] = momamps_kperp_temp[imom]
            self.momamps_kperp[mom].generate_timeseries()
            self.momamps_kperp[mom].generate_timeaverage()



class FluxSpectra(object):
    """ kx and ky radial flux spectra

    Collect and calculate the kx and ky spectra for the radial fluxes from
    mom files. For further information see diagnostics/doc/spectra.pdf

    """

    def __init__(self, common, species, rundatafiles, with_momentum=True):
        if not common.y_local:
            raise NotImplementedError("Not implemented for y-global simulations yet")
        self.cm = common
        self.species = species
        self.timearray = None
        self.fluxes_kx = {"Gammaes": [], "Qes": []}
        self.fluxes_ky = {"Gammaes": [], "Qes": []}
        if with_momentum:
            self.fluxes_kx.update({"Pies": []})
            self.fluxes_ky.update({"Pies": []})
        if common.electromagnetic:
            self.fluxes_kx.update({"Gammaem": [], "Qem": []})
            self.fluxes_ky.update({"Gammaem": [], "Qem": []})
            if with_momentum:
                self.fluxes_kx.update({"Piem": []})
                self.fluxes_ky.update({"Piem": []})
        self.calc_fluxes(species=species, rundatafiles=rundatafiles, with_momentum=with_momentum)

    def _fetch_moms(self, rundatafiles, species):
        """ Get mom and field slice object for species"""
        diagspace = DiagSpace(self.cm.spatialgrid, x_fourier=True, y_fourier=True)
        momamps = {"phi": None, "dens": None, "tpar": None, "tperp": None, "qpar": None,
                   "qperp": None, "upar": None}
        if self.cm.electromagnetic:
            momamps.update({"apar": None})
            if self.cm.bpar:
                momamps.update({"bpar": None})
        with Pool(len(momamps)) as p:
            momamp_temp = p.map(
                lambda mom: MomFieldSlice(self.cm, mom, species, diagspace, rundatafiles,
                                          modifier_func=lambda dum: dum), momamps)
        for imom, mom in enumerate(momamps):
            momamps[mom] = momamp_temp[imom]
        return momamps

    def calc_fluxes(self, species, rundatafiles, with_momentum):
        """ Calculate the flux spectra"""
        momamps = self._fetch_moms(rundatafiles, species=species)
        ispec = self.cm.specnames.index(species) + 1
        geom = pydiag.utils.geom.Geometry(self.cm)
        momamps["dens"].check_times()
        pos = momamps["dens"].calc_positions()
        self.timearray = np.take(momamps["dens"].timearray, pos)
        for time in self.timearray:
            # vE_x = -c/C_{xy} d phi / dy
            vE_x = - 1j*self.cm.spatialgrid.ky[np.newaxis, :, np.newaxis]*momamps[
                "phi"].generate_slice_attime(time)
            if self.cm.electromagnetic:
                # B_x = d A_par / dy
                B_x = 1j*self.cm.spatialgrid.ky[np.newaxis, :, np.newaxis]*momamps[
                    "apar"].generate_slice_attime(time)
                if self.cm.bpar:
                    # (d B_par / dy) / B0(z)
                    dBpar_dy = 1j*1/self.cm.pnt.Bref*self.cm.spatialgrid.ky[np.newaxis, :,
                                                     np.newaxis]*momamps[
                                   "apar"].generate_slice_attime(time)
                    dBpar_dy /= geom.Bfield[np.newaxis, np.newaxis, :]
                else:
                    dBpar_dy = 0

            def mom_t(momname):
                return momamps[momname].generate_slice_attime(time)

            # Required due to normalisation
            n0 = self.cm.pars["dens{}".format(ispec)]
            mass = self.cm.pars["mass{}".format(ispec)]
            T0 = self.cm.pars["temp{}".format(ispec)]
            # Gamma_es = <n * ve_x>*n0 (n0 due to normalisation)
            # Q_es = (1/2 Tpar + Tperp + 3/2 n) ve_x
            fluxfactors = {"Gammaes": vE_x*np.conj(mom_t("dens"))*n0, "Qes": vE_x*np.conj(
                    0.5*mom_t("tpar") + mom_t("tperp") + 1.5*mom_t("dens"))*n0*T0}
            if with_momentum:
                warnings.warn("The momentum transport spectra are still incorrect", RuntimeWarning)
                #  P_es = <upar ve_x>
                fluxfactors.update({"Pies": vE_x*mom_t("upar")*n0*mass})
            if self.cm.electromagnetic:
                # Gamma_em = <upar * B_x>
                # Q_em = (qpar + qperp + 5/2 upar) B_x
                # "qpar" and "qperp" from mom file are NOT identical to qpar and qperp
                # but to qpar+1.5upar and qperp+upar
                fluxfactors.update({"Gammaem": B_x*np.conj(mom_t("upar"))*n0,
                                    "Qem": B_x*np.conj(mom_t("qpar") + mom_t("qperp"))*n0*T0})
                if with_momentum:
                    # P_em = (Tpar + n) B_x
                    warnings.warn("The momentum transport spectra are still incorrect",
                                  RuntimeWarning)
                    fluxfactors.update({"Piem": B_x*(mom_t("tpar") + mom_t("dens"))*n0*T0})
                if self.cm.bpar:
                    raise NotImplementedError("bpar part of the fluxes is not implemented yet")
                    fluxfactors["Gammaem"] += 0
                    fluxfactors["Qem"] += 0
            for flux in self.fluxes_kx:
                self.fluxes_kx[flux].append(averages.flux_spectra_yz_av(fluxfactors[flux], geom))
                self.fluxes_ky[flux].append(averages.flux_spectra_xz_av(fluxfactors[flux], geom))
        for flux in self.fluxes_kx:
            self.fluxes_kx[flux] = averages.mytrapz(self.fluxes_kx[flux], self.timearray)
            self.fluxes_ky[flux] = averages.mytrapz(self.fluxes_ky[flux], self.timearray)

    def print_total_fluxes(self):
        """Print the total flux for comparison with nrg diag"""
        print("============================")
        print("Total flux for species {}:".format(self.species))
        for flux in self.fluxes_kx:
            print("{} integrated kx spectrum: {}".format(flux, np.sum(self.fluxes_kx[flux])))
            print("{} integrated ky spectrum: {}".format(flux, np.sum(self.fluxes_ky[flux])))
        print("\n")


class PlotAmplitudeSpectra(Plotting):

    def __init__(self, commonlist, spectraserieslist):
        super().__init__()
        self.spectralist = spectraserieslist
        self.commonlist = commonlist


    def createfigs(self):
        for spectraseries in self.spectralist:
            moms = spectraseries.momamps_kx.keys()
            fig_kx = plt.figure()
            fig_ky = plt.figure()

            ax_log_kx = fig_kx.add_subplot(2, 1, 1)
            ax_lin_kx = fig_kx.add_subplot(2, 1, 2)
            ax_log_ky = fig_ky.add_subplot(2, 1, 1)
            ax_lin_ky = fig_ky.add_subplot(2, 1, 2)
            kx = spectraseries.cm.spatialgrid.kx
            ky = spectraseries.cm.spatialgrid.ky
            for mom in moms:
                ax_log_kx.plot(kx, np.fft.fftshift(spectraseries.momamps_kx[mom].timeaverage),
                               label=self.titles[mom])
                ax_lin_kx.plot(kx, np.fft.fftshift(spectraseries.momamps_kx[mom].timeaverage),
                               label=self.titles[mom])
                ax_log_ky.plot(ky, spectraseries.momamps_ky[mom].timeaverage,
                               label=self.titles[mom])
                ax_lin_ky.plot(ky, spectraseries.momamps_ky[mom].timeaverage,
                               label=self.titles[mom])
            ax_log_kx.loglog()
            ax_log_ky.loglog()
            ax_log_kx.set_xlabel(r"$k_x \rho_{ref}$")
            ax_lin_kx.set_xlabel(r"$k_x \rho_{ref}$")
            ax_log_ky.set_xlabel(r"$k_y \rho_{ref}$")
            ax_lin_ky.set_xlabel(r"$k_y \rho_{ref}$")
            for ax in [ax_log_kx, ax_lin_kx, ax_lin_ky, ax_log_ky]:
                ax.set_ylabel(r"$<|A|^2>$")
                ax.legend()
            ax_log_kx.set_title("{}".format(spectraseries.species))
            ax_log_ky.set_title("{}".format(spectraseries.species))

class PlotPerpSpectra(Plotting):

    def __init__(self, commonlist, spectraserieslist):
        super().__init__()
        self.spectralist = spectraserieslist
        self.commonlist = commonlist
        self.titles.update({"phi": r"$E_\perp$", "apar": r"$B_\perp$", "dens": r"$n$",
                            "tpar": r"$T_\parallel$", "tperp": r"$T_\perp$",
                            "qpar": r"$q_\parallel + 1.5p_0 u_\parallel$",
                            "qperp": r"$q_\perp + p_0 u_\parallel$", "upar": r"$u_\parallel$",
                            "bpar": r"$B_\parallel$", "n00": r"$N_{00}$",
                            "n20": r"$N_{20}$", "n02": r"$N_{02}$"})

    def createfigs(self, moms=None):
        for spectraseries in self.spectralist:
            if not moms:
                moms = spectraseries.momamps_kperp
            fig_kperp = plt.figure()

            ax_log_kperp = fig_kperp.add_subplot(1, 1, 1)

            kx = spectraseries.cm.spatialgrid.kx_fftorder
            ky = spectraseries.cm.spatialgrid.ky
            kperp_sq_2d = kx[:,None]**2+ky[None,:]**2

            #generate kperp array for binning, default number of bins is 1.5*nky0
            nbins=int(np.rint(0.5*len(ky)))
            kperp_bins=np.linspace(0,np.sqrt(np.amax(kperp_sq_2d)),nbins+1)
            #kperp bin centers for plotting (match also array sizes)
            kperp_bc=kperp_bins[:-1]+0.5*np.diff(kperp_bins)[0]

            #logarithmic binning -- disabled by default
            #kperp_bins=np.logspace(np.log10(0.5*ky[1]),np.log10(np.sqrt(np.amax(kperp_sq_2d))),nbins+1)
            #bin "centers" as geometric means of bin boundaries
            #kperp_bc=np.sqrt(kperp_bins[:-1]*kperp_bins[1:])

            for mom in moms:
                if mom in ("phi","apar"):
                    tmp=kperp_sq_2d*spectraseries.momamps_kperp[mom].timeaverage
                else:
                    tmp=spectraseries.momamps_kperp[mom].timeaverage
                #generate 1d perpendicular spectra by binning 2d kperp values
                #with spectral amplitude as the weight
                mom_kperp,kperp_bins=np.histogram(np.sqrt(kperp_sq_2d),bins=kperp_bins,weights=tmp)
                ax_log_kperp.plot(kperp_bc, mom_kperp,label=self.titles[mom])
            xmin,xmax=kperp_bc[0],kperp_bc[np.int(nbins/np.sqrt(2))]
            ax_log_kperp.set_xlim(xmin,xmax)
            kp_mhd=np.linspace(xmin,1,128)
            kp_kin=np.linspace(1,xmax,128)
            ax_log_kperp.plot(kp_mhd, kp_mhd**(-5./3),ls='--',lw=1,label="$k_\perp^{-5/3}$")
            ax_log_kperp.plot(kp_kin, kp_kin**(-2.8),ls='--',lw=1,label="$k_\perp^{-2.8}$")
            #ax_log_kperp.plot(kperp_bc, kperp_bc**(-7./3),ls='--',lw=1,label="$k_\perp^{-7/3}$")
            ax_log_kperp.loglog()
            ax_log_kperp.set_xlabel(r"$k_\perp \rho_{ref}$")
            ax_log_kperp.set_ylabel(r"$\langle|A|^2\rangle$")
            ax_log_kperp.legend()
            ax_log_kperp.set_title("{}".format('Field energy spectra'))
            fig_kperp.tight_layout()


class PlotFluxSpectra(Plotting):

    def __init__(self, commonlist, fluxspecserieslist):
        super().__init__()
        self.fluxspeclist = fluxspecserieslist
        self.commonlist = commonlist
        self.titles.update({"Gammaes": r"$\Gamma_{es}$", "Qes": r"$Q_{es}$", "Pies": r"$\Pi_{es}$",
                            "Gammaem": r"$\Gamma_{em}$", "Qem": r"$Q_{em}$", "Piem": r"$\Pi_{em}$"})

    def createfigs(self):
        for fluxes in self.fluxspeclist:
            fig_kx = plt.figure(figsize=(6, 8))
            fig_ky = plt.figure(figsize=(6, 8))

            ax_log_kx = fig_kx.add_subplot(2, 1, 1)
            ax_lin_kx = fig_kx.add_subplot(2, 1, 2)
            ax_log_ky = fig_ky.add_subplot(2, 1, 1)
            ax_lin_ky = fig_ky.add_subplot(2, 1, 2)
            kx = fluxes.cm.spatialgrid.kx_pos
            ky = fluxes.cm.spatialgrid.ky
            for flux in fluxes.fluxes_kx:
                # Mask negative flux values for solid lines
                pos_flux_kx = np.ma.masked_where((fluxes.fluxes_kx[flux] <= 0),
                                                 fluxes.fluxes_kx[flux])
                # Mask zero flux for dashed lines, this takes care of the Nyquist mode in kx
                all_flux_kx = np.ma.masked_where((fluxes.fluxes_kx[flux] == 0),
                                                 fluxes.fluxes_kx[flux])
                pos_flux_ky = np.ma.masked_where((fluxes.fluxes_ky[flux] <= 0),
                                                 fluxes.fluxes_ky[flux])
                all_flux_ky = np.ma.masked_where((fluxes.fluxes_ky[flux] == 0),
                                                 fluxes.fluxes_ky[flux])
                baselogkx, = ax_log_kx.plot(kx, pos_flux_kx, label=self.titles[flux])
                ax_log_kx.plot(kx, np.abs(all_flux_kx), ls="--", color=baselogkx.get_color())
                ax_lin_kx.plot(kx, all_flux_kx, label=self.titles[flux])
                baselogky, = ax_log_ky.plot(ky, pos_flux_ky, label=self.titles[flux])
                ax_log_ky.plot(ky, np.abs(all_flux_ky), ls="--", color=baselogky.get_color())
                ax_lin_ky.plot(ky, all_flux_ky, label=self.titles[flux])
                #np.savez("fluxspectra_{}_{}".format(fluxes.species, flux), ky=ky,
                #         fluxspectrum=all_flux_ky)
            ax_log_kx.loglog()
            ax_log_ky.loglog()
            ax_lin_kx.set_xlim(left=0)
            ax_log_kx.set_xlabel(r"$k_x \rho_{ref}$")
            ax_lin_kx.set_xlabel(r"$k_x \rho_{ref}$")
            ax_log_ky.set_xlabel(r"$k_y \rho_{ref}$")
            ax_lin_ky.set_xlabel(r"$k_y \rho_{ref}$")
            for ax in [ax_log_kx, ax_lin_kx, ax_lin_ky, ax_log_ky]:
                # ax.set_ylabel(r"$<|A|^2>$")
                ax.legend()
            ax_log_ky.set_title("{}".format(fluxes.species))
            ax_log_kx.set_title("{}".format(fluxes.species))
            fig_kx.tight_layout()
            fig_ky.tight_layout()



""" profile_data.py

Contains the infrastructure to handle profile_spec_fileextension files
Also includes nustar and the Chang-Hinton neoclassical heat flux prediction
"""

import math
import numpy as np

from .base_file import TimeSeries
from ..utils.geom import Geometry
from ..utils.nc_predictions import chang_hinton


class ProfileFile(TimeSeries):
    """Class to handle a profile diagnostic file

    :param filename: Typically profile_{spec}{fileextension}
    :param common: CommonData object of the run
    """

    def __init__(self, filename, common):
        super().__init__(filename, common)
        self.dataarray = {"Ts": [], "ns": [], "omts": [], "omns": [], "Gammanc": [], "Gammaturb": [],
                         "Qturb": [], "Qnc": [], "Pinc": [], "Piturb": [], "jbs": []}
        # Collisionality, i.e. ratio of \nu and bounce frequency
        self.nustar = []

    def generate_timeseries(self):
        """ Read density, temperature, gradients and fluxes """
        self.timearray, blocks = self._parse_profile_file()
        self.timearray = np.array(self.timearray)
        self.check_times()
        #  For a single time, find element closest to given input time
        pos = self.calc_positions()
        self._addprofdata(blocks, pos)
        self.timearray = self.timearray[pos]

    def _parse_profile_file(self):
        blocks = []
        timearray = []
        print('Reading  {}\n'.format(self.objectname))
        try:
            with open(self.objectname) as prfile:
                next(prfile)  # skip header
                for line in prfile:
                    if not line or line.startswith('\n'):
                        continue
                    if line.startswith('#'):
                        timearray.append(float(line.split()[1]))
                        blocks.append([])
                    else:
                        blocks[-1].append(line)
        except IOError:
            print("Probably profile file does not exist: {}".format(self.objectname))
            raise
        return timearray, blocks

    def _addprofdata(self, blocks, pos):
        """ Fill vectors with profile data """
        for entry in pos:
            self.dataarray["Ts"].append(np.zeros(self.cm.pnt.nx0))
            self.dataarray["ns"].append(np.zeros(self.cm.pnt.nx0))
            self.dataarray["omts"].append(np.zeros(self.cm.pnt.nx0))
            self.dataarray["omns"].append(np.zeros(self.cm.pnt.nx0))
            self.dataarray["Gammaturb"].append(np.zeros(self.cm.pnt.nx0))
            self.dataarray["Qturb"].append(np.zeros(self.cm.pnt.nx0))
            self.dataarray["Piturb"].append(np.zeros(self.cm.pnt.nx0))
            self.dataarray["Gammanc"].append(np.zeros(self.cm.pnt.nx0))
            self.dataarray["Qnc"].append(np.zeros(self.cm.pnt.nx0))
            self.dataarray["Pinc"].append(np.zeros(self.cm.pnt.nx0))
            self.dataarray["jbs"].append(np.zeros(self.cm.pnt.nx0))
            for i in range(0, self.cm.pnt.nx0):
                linearr = (blocks[entry][i]).split()
                self.dataarray["Ts"][-1][i] = linearr[2]
                self.dataarray["ns"][-1][i] = linearr[3]
                self.dataarray["omts"][-1][i] = linearr[4]
                self.dataarray["omns"][-1][i] = linearr[5]
                self.dataarray["Gammaturb"][-1][i] = linearr[6]
                self.dataarray["Qturb"][-1][i] = linearr[7]
                self.dataarray["Piturb"][-1][i] = linearr[8]
                self.dataarray["Gammanc"][-1][i] = linearr[9]
                self.dataarray["Qnc"][-1][i] = linearr[10]
                self.dataarray["Pinc"][-1][i] = linearr[11]
                self.dataarray["jbs"][-1][i] = linearr[12]
            # Correct for convective energy flux
            self.dataarray["Qnc"][-1] -= 2.5*self.dataarray["Gammanc"][-1]*self.dataarray["Ts"][-1]
            self.dataarray["Qturb"][-1] -= 2.5*self.dataarray["Gammaturb"][-1]*self.dataarray["Ts"][-1]


class ProfileData(object):
    """Class handling the profile diagnostic files of a run (i.e. one per species)

    :param common: CommonData object of the run
    """

    def __init__(self, common, rundatafiles):
        """ Read parameter file and create empty arrays for profile data """
        self.isdatapresent = False
        self.cm = common
        self.rundatafiles = rundatafiles
        self.geom = Geometry(common)
        self.q = np.array(self.geom.q)
        self.prspec = []
        self.xs = np.empty(self.cm.pnt.nx0)
        self.T0s = np.empty((self.cm.pnt.nx0, self.cm.pnt.n_spec))
        self.n0s = np.empty((self.cm.pnt.nx0, self.cm.pnt.n_spec))
        self.omt0s = np.empty((self.cm.pnt.nx0, self.cm.pnt.n_spec))
        self.omn0s = np.empty((self.cm.pnt.nx0, self.cm.pnt.n_spec))

        self.Afs = []  # Flux surface area
        # Chang-Hinton prediction for main ion species
        self.Qcharr = []

    def calc_afs(self):
        """ Calculate area of flux surfaces"""
        c = self.cm.pnt.ly*self.cm.pnt.rhostar*self.cm.pnt.minor_r*2.0*np.pi*self.cm.pnt\
            .n0_global/self.cm.pnt.nz0
        self.Afs = c*np.sum(self.geom.jacobian*np.sqrt(self.geom.gxx), axis=0)/self.cm.pnt.major_R

    def _read_zero_prof(self):
        """Get the initial profiles from profiles_spec_fileextension"""
        for n in range(self.cm.pnt.n_spec):
            prof0file = open('profiles_{}{}'.format(self.cm.specnames[n], self.cm.fileextension))
            lines = prof0file.readlines()
            for i in range(0, self.cm.pnt.nx0):
                l = 2 + i
                self.xs[i] = lines[l].split()[0]
                self.T0s[i, n] = float(lines[l].split()[2]) / self.cm.pnt.Tref
                self.n0s[i, n] = float(lines[l].split()[3]) / self.cm.pnt.nref
                self.omt0s[i, n] = lines[l].split()[4]
                self.omn0s[i, n] = lines[l].split()[5]
            prof0file.close()

    def get_profiles(self):
        """ Organize the profile data, calls most other functions"""
        self._read_zero_prof()
        for n in range(0, self.cm.pnt.n_spec):
            prs = self.rundatafiles.get_fileobject('profile_{}'.format(self.cm.specnames[n]))
            prs.generate_timeseries()
            self.prspec.append(prs)
        self.calc_nustar()
        self.Qcharr = chang_hinton(self, 0)  # Only main ion species
        self.calc_afs()
        self.isdatapresent = True

    def calc_nustar(self):
        """ Calculate the collisionality nustar for each species. """
        try:
            coll = self.cm.pnt.coll
        except AttributeError:
            coll = 0
        epsilon = np.array(self.xs * self.cm.pnt.minor_r / self.cm.pnt.major_R)
        for prs in self.prspec:
            prs.nustar = []
            for n, T in zip(prs.dataarray["ns"], prs.dataarray["Ts"]):
                prs.nustar.append(np.array(8.0 / 3.0 / math.sqrt(math.pi) *
                                           self.cm.pnt.major_R * coll / epsilon ** (3.0 / 2) *
                                           self.q * n / T ** 2))

    def generate_timetrace(self, xpos, quantity="Qturb"):
        """ Generate a 1d-timetrace of a profile quantity.

        :param xpos: Radial position in x/a, the nearest x gridpoint will be used.
        :param quantity: Determines the quantity which will be used, default turbulent heat flux
        """
        xind = (np.abs(self.xs - xpos)).argmin()  # find index of position closest to xpos
        timetr = []
        for n, prs in enumerate(self.prspec):
            for step in prs.dataarray[quantity]:
                timetr.append(step[xind])
            timetrfile = open("ttrace_{}{}_{}{}".format(xpos, quantity, self.cm.specnames[n],
                                                        self.cm.fileextension), mode='w')
            timetrfile.write("# time  " + quantity + "\n")
            for t, q in zip(prs.timearray, timetr):
                timetrfile.write(str(t) + "  " + str(q) + "\n")
            timetrfile.close()


def glueprofiledata(prdlist):
    """ Take profdata objects with matching time intervals and combine them into a larger one
    :param prdlist: List of ProfileData objects to combine
    :returns: Combined Profdata object

    """
    # TODO: sanity checks for the parameters of prdlist entries
    # Nothing to combine, nothing to do
    if len(prdlist) == 1:
        return prdlist[0]
    # Use first ProfileData as basis for the construction of the glued object
    result = prdlist.pop(0)
    for prd in prdlist:
        for n in range(result.cm.pnt.n_spec):
            # When times overlap, give following ProfileData preference
            # If the timefield has been converted to a np array before we need to undo it
            result.prspec[n].timearray = result.prspec[n].timearray.tolist()
            prd.prspec[n].timearray = prd.prspec[n].timearray.tolist()

            while result.prspec[n].timearray[-1] >= prd.prspec[n].timearray[0]:
                del result.prspec[n].timearray[-1]
                for key in result.prspec[n].dataarray:
                    del result.prspec[n].dataarray[key][-1]
            result.prspec[n].timearray = np.concatenate((result.prspec[n].timearray,
                                                         prd.prspec[n].timearray))
            for key in result.prspec[n].dataarray:
                result.prspec[n].dataarray[key] += prd.prspec[n].dataarray[key]
    result.cm.endtime = prdlist[-1].cm.endtime
    result.calc_nustar()
    result.Qcharr = chang_hinton(result, 0)
    del prdlist
    return result



# -*- coding: utf-8 -*-
"""
Created on Tue Jan  7 14:39:26 2020

@author: dykua, venkitesh ayyar

For downloading files from mgk_fusion in shell
"""

import sys
import os
import argparse
from pymongo import MongoClient
from bson.objectid import ObjectId
import gridfs

from mgkdb.support.mgk_file_handling import get_oid_from_query, Str2Query, download_dir_by_name, download_file_by_path, download_file_by_id, download_runs_by_id
from mgkdb.support.mgk_login import mgk_login,f_login_dbase

def f_parse_args():
    #==========================================================
    # argument parser
    #==========================================================
    parser = argparse.ArgumentParser(description='Process input for downloading files')

    parser.add_argument('-Q', '--query', default= None,help='mongodb query')
    parser.add_argument('-T', '--target', default= None,help='run collection_name, i.e. gene output folder path')
    parser.add_argument('-F', '--file', default = None, help='filename to be downloaded if any')
    parser.add_argument('-C', '--collection', choices=['linear','nonlinear','files'], default='linear', type=str, help='collection name in the database')
    parser.add_argument('-OID', '--objectID', default = None, help = 'Object ID in the database')
    parser.add_argument('-A', '--authenticate', default = None, help='locally saved login info, a .pkl file')
    parser.add_argument('-D', '--destination', default = '', help = 'directory where files are downloaded to.')
    parser.add_argument('-S', '--saveas', default = 'specific_file', help = 'Name to save the file as')

    return parser.parse_args()

### Main 
def main_download(target, file, objectID, destination, saveas, query, authenticate, collection):

    OID = objectID
    op_fname = saveas


    ### Connect to database 
    login = f_login_dbase(authenticate)
    client, database = login.connect()
    with client:
        ## Dict to convert from argument to collection name in database
        collection_dict={'linear':'LinearRuns','nonlinear':'NonlinRuns','files':'fs.files'}
        collection_name =  getattr(database,collection_dict[collection])

        if query:
            print("working on query: {} ......".format(query))
            found = get_oid_from_query(database, collection_name, Str2Query(query))
            for oid in found:
                download_runs_by_id(database, collection_name, oid, destination)

        elif file:
            download_file_by_path(database, file, destination, revision=-1, session=None)   
            
        elif OID:
            if collection=='files': 
                download_file_by_id(database, ObjectId(OID), destination, op_fname, session = None)
            elif collection in ['linear','nonlinear']:
                download_runs_by_id(database, collection_name, ObjectId(OID), destination)
            else : 
                print("Invalid option for collection for OID",collection)
                raise SystemError
        elif target:
            download_dir_by_name(database, collection_name, target, destination)


def main():

    ### Parse arguments 
    args = f_parse_args()
    # print(args)

    main_download(**vars(args))


if __name__=="__main__":
    main()




from os.path import getsize

import numpy as np
from .base_file import BinaryFile


class FieldFile(BinaryFile):
    """ Class to parse binary field.dat diagnostic output of GENE

    :param filename: Should be "field.dat" or "field_fileextension"
    :param common: CommonData object of the run to analyse
    """
    def __init__(self, filename, common, usemmap=False):
        super().__init__(filename, common, usemmap=usemmap)
        self.ftind = [-1]*self.nfields
        self.fieldvar3d = {"phi": None, "apar": None, "bpar": None}
        self.phi3d = None
        self.apar3d = None
        self.bpar3d = None

    def _set_sizes(self):
        super()._set_sizes()
        self.nfields = int(self.cm.pars['n_fields'])
        self.entrysize = self.cm.pnt.nx0*self.cm.pnt.nky0*self.cm.pnt.nz0*self.complexsize
        # jump in bytes in field files
        self.leapfld = self.nfields*(self.entrysize + 2*self.intsize)

    def get_timearray(self):
        """Get time array for field file """
        self.timearray = []
        self.file.seek(0)
        for i in range(int(getsize(self.filename)/(self.leapfld + self.tesize))):
            self.timearray.append(float(self.te.unpack(self.file.read(self.tesize))[1]))
            self.file.seek(self.leapfld, 1)

    def offset(self, var):
        """Calculate offset in field file for a given self.time and variable"""
        if var in [i for i in range(self.nfields)]:
            return self.tesize + self.tind*(self.tesize + self.leapfld) + var*(
                self.entrysize + 2*self.intsize) + self.intsize

    def readvar(self, var):
        """ Return 3d field data at the time set in self.time"""
        var3d = self.readfunc(self.file, count=self.cm.pnt.nx0*self.cm.pnt.nky0*self.cm.pnt.nz0,
                              dtype=self.npct, offset=self.offset(var))

        # Bring array into x, y. z order
        if self.cm.x_local and not self.cm.y_local:  # y-global has yx order
            var3d = var3d.reshape(self.cm.pnt.nky0, self.cm.pnt.nx0, self.cm.pnt.nz0, order="F")
            var3d = np.swapaxes(var3d, 0, 1)
        else:
            var3d = var3d.reshape(self.cm.pnt.nx0, self.cm.pnt.nky0, self.cm.pnt.nz0, order="F")
        return var3d

    def reset_tinds(self):
        """ Reset time indices when reloading file"""
        super().reset_tinds()
        self.ftind = [-1]*self.nfields

    def get_var(self, varname):
        """ """
        varidx = {"phi": 0, "apar": 1, "bpar": 2}
        if varidx[varname] < self.nfields:
            if not self.ftind[varidx[varname]] == self.tind:
                self.ftind[varidx[varname]] = self.tind
                self.fieldvar3d[varname] = self.readvar(varidx[varname])
        return self.fieldvar3d[varname]

    def phi(self):
        """ Return phi (electrostatic potential)

        Read from file if necessary
        """
        if not self.ftind[0] == self.tind:
            self.ftind[0] = self.tind
            self.phi3d = self.readvar(0)
        return self.phi3d

    def apar(self):
        """ Return apar (parallel component of magnetic vector potential)

        Read from file if necessary
        """
        if self.nfields > 1:
            if not self.ftind[1] == self.tind:
                self.ftind[1] = self.tind
                self.apar3d = self.readvar(1)
            return self.apar3d

    def bpar(self):
        """ Return bpar (parallel component of magnetic field fluctuation)

        Read from file if necessary
        """
        if self.nfields > 2:
            if not self.ftind[2] == self.tind:
                self.ftind[2] = self.tind
                self.bpar3d = self.readvar(2)
            return self.bpar3d



""" gkdb.py: class handling data conversion and writing to the GyroKineticDataBase (https://github.com/gkdb/gkdb) """
import re
import os
import datetime
import numpy as np
import warnings
from collections import OrderedDict
from  .geom import Geometry
# import .comm
from ..data.omega_eigenvalue import Eigenvaluedata
from ..diagplots.plot_ball import ModeStructure

import json
#import .averages as averages
import matplotlib.pyplot as plt #remove later

class GKDB_linear(object):
    """ GKDB class:

    Converts the results of a GENE run into a GKDB-style json file

    Currently restricted to rotationless linear fluxtube simulations"
    """

    def __init__(self,common,rundatafiles,creator):
        # dictionary for gkdb parameters: {parameter: value}
        self.creator = creator
        self.cm = common
#        self.precheck()
        self.geom = Geometry(self.cm)
        self.fields = ["phi"]
        if common.electromagnetic:
            self.fields += ["apar"]
            if common.bpar:
                self.fields += ["bpar"]

        self.rundatafiles = rundatafiles
        self.gkdict = OrderedDict()
        self.set_gkdb_shape()
        self.set_gene2gkdb_conversion()
        self.fill_gkdb_from_par()
        

    def precheck(self):
        """ Run some basic checks before actual execution """
        if self.cm.nonlinear:
            raise RuntimeError("Nonlinear simulations not yet supported")
        else:
            if self.cm.pars['comp_type']=="'IV'":
                if self.cm.starttime!=-2 or self.cm.endtime!=-2:
                    print("*** WARNING: Selecting l:l as time window could be more appropriate ***")
            else:
                if (self.cm.starttime!=-1 and self.cm.starttime!=1) or \
                   (self.cm.endtime!=-2 and self.cm.endtime!=self.cm.pars["n_ev"]):
                    raise RuntimeError("Select f:l as time window for EV simulations")
        if self.cm.pars['magn_geometry']!="'miller'":
            raise RuntimeError("Only Miller geometries supported for now")


    def set_gkdb_shape(self):
        """ Determines shape parameters for GKDB """

        print ("Determining flux surface shape ...")
        nz0 = self.cm.pnt.nz0

        #determine flux surface center
        self.R0 = 0.5*(max(self.geom.R)+min(self.geom.R))
        self.Z0 = 0.5*(max(self.geom.Z)+min(self.geom.Z))
        print ("(R0,Z0) = (", self.R0, ",", self.Z0,")")
#        self.R0 = np.sum(self.geom.R*self.geom.jacobian)/np.sum(self.geom.jacobian)
#        self.Z0 = np.sum(self.geom.Z*self.geom.jacobian)/np.sum(self.geom.jacobian)

        #minor radius of flux surface
        self.r = 0.5*(max(self.geom.R)-min(self.geom.R))

        #poloidal theta (COCOS=11)
        self.theta = np.arctan(-(self.geom.Z-self.Z0)/(self.geom.R-self.R0))
        for iz in range(int(nz0/2),nz0-1):
            thetadiff = (self.theta[iz+1]-self.theta[iz])
            if thetadiff>0.5*np.pi:
                self.theta[iz+1]-=np.pi
            elif thetadiff<-0.5*np.pi:
                self.theta[iz+1]+=np.pi
            thetadiff = (self.theta[nz0-1-iz-1]-self.theta[nz0-1-iz])
            if thetadiff>0.5*np.pi:
                self.theta[nz0-1-iz-1]-=np.pi
            elif thetadiff<-0.5*np.pi:
                self.theta[nz0-1-iz-1]+=np.pi

        #sorted theta (starting at th=0)
        theta_sort = (self.theta+2.0*np.pi)%(2.0*np.pi)
        sort_ind = np.argsort(theta_sort)
        theta_sort = theta_sort[sort_ind]

        #Flux surface distance (original grid and normalized)
        aN = np.sqrt((self.geom.R[sort_ind]-self.R0)**2+(self.geom.Z[sort_ind]-self.Z0)**2)/self.R0
        aN_dr = ((self.geom.R[sort_ind]-self.R0)*self.geom.dxdR[sort_ind]+
                 (self.geom.Z[sort_ind]-self.Z0)*self.geom.dxdZ[sort_ind])/\
            (aN*self.geom.gxx[sort_ind])/self.R0

        #equidistant theta (COCOS=11) with somewhat higher resolution for fourier transforms etc
        nth = 4*nz0
        theta_grid = (np.arange(0.0,nth,1.0))*2.0*np.pi/nth

        #interpolate to this equidist. grid with proper order (interpol requires increasing arr)
        aN_th = np.interp(theta_grid,theta_sort,aN,period=2.0*np.pi)
        aN_dr_th = np.interp(theta_grid,theta_sort,aN_dr,period=2.0*np.pi)

        #now fourier transform
        fft_aN = np.fft.fft(aN_th)
        fft_aN_dr = np.fft.fft(aN_dr_th)

        #keep up to nth_kept modes (typically around 8)
        nth_kept = 8 #nz0
        self.cN = 2.0*np.real(fft_aN)[0:nth_kept]/nth
        self.cN[0] *= 0.5
        self.sN = -2.0*np.imag(fft_aN)[0:nth_kept]/nth
        self.sN[0] *= 0.5
        self.cN_dr = 2.0*np.real(fft_aN_dr)[0:nth_kept]/nth
        self.cN_dr[0] *= 0.5
        self.sN_dr = -2.0*np.imag(fft_aN_dr)[0:nth_kept]/nth
        self.sN_dr[0] *= 0.5

        #check parametrization
        nind = np.arange(0.0,nth_kept,1.0)
        a_check = []
        a_dr_check = []
        for iz in range(0,nth):
            a_check += [np.sum(self.cN*np.cos(nind*theta_grid[iz])+
                               self.sN*np.sin(nind*theta_grid[iz]))]
            a_dr_check += [np.sum(self.cN_dr*np.cos(nind*theta_grid[iz])+
                                  self.sN_dr*np.sin(nind*theta_grid[iz]))]

        a_dr_check = np.array(a_dr_check)
        err = np.sum(np.abs(a_check-aN_th)/aN_th)/nth
        print ("Relative error of flux surface parametrization: {0:12.6E}".format(err))

        #some plot diagnostics
        show_aN = False
        if show_aN:
            plt.plot(theta_sort,aN,label='aN_orig')
            plt.plot(theta_grid,aN_th,label='aN_th,intermediate')
            plt.plot(theta_grid,a_check,label='aN_final')
            plt.plot(theta_sort,aN_dr,label='aN_dr_orig')
            plt.plot(theta_grid,a_dr_check,label='aN_dr_final')
            plt.legend()
            plt.show()

        show_flux_surface = False
        if show_flux_surface:
            dr = 0.04  #radial extension for testing

            plt.plot(self.geom.R[sort_ind],self.geom.Z[sort_ind],'o',label="original")
            plt.plot((self.geom.R[sort_ind]+dr*self.geom.dxdR[sort_ind]/self.geom.gxx[sort_ind]),
                     (self.geom.Z[sort_ind]+dr*self.geom.dxdZ[sort_ind]/self.geom.gxx[sort_ind]),
                     'x',label="original+dr")
            theta_grid = (np.arange(0.0,nth,1.0))*2.0*np.pi/(nth-1)
            # R_n_th = self.R0+aN_th*np.cos(theta_grid)
            # Z_n_th = self.Z0-aN_th*np.sin(theta_grid)
            # plt.plot(R_n_th,Z_n_th,label="intermediate")
            R_check = self.R0+a_check*np.cos(theta_grid)*self.R0
            Z_check = self.Z0-a_check*np.sin(theta_grid)*self.R0
            plt.plot(R_check,Z_check,label="gkdb")
            R_check = self.R0+(a_check+dr/self.R0*a_dr_check)*np.cos(theta_grid)*self.R0
            Z_check = self.Z0-(a_check+dr/self.R0*a_dr_check)*np.sin(theta_grid)*self.R0
            plt.plot(R_check,Z_check,label="gkdb+dr")
            plt.xlabel('R/m')
            plt.ylabel('Z/m')
            plt.axis('equal')
            plt.legend()
            plt.show()

    def set_gene2gkdb_conversion(self):
        """ set conversion factors for various quantities """

        print ("Setting conversion factors ...")

        #determine index of electron species
        ispec_electrons = -1111
        for spec in self.cm.specnames:
            ispec = self.cm.specnames.index(spec) + 1
            if self.cm.pars["charge{}".format(ispec)]==-1:
                ispec_electrons = ispec

        self.gene2gkdb = OrderedDict()
        self.gene2gkdb["Lref"] = self.cm.pnt.Lref/self.R0 #Lref_GENE/Lref_GKDB
        self.gene2gkdb["Bref"] = self.cm.pnt.Lref*self.cm.pnt.major_R/self.R0  #JUST 1/R correction for the moment
        print ("Bref_GENE/Bref_GKDB = ", self.gene2gkdb["Bref"])

        #CHECK THIS: x dimensionless, r dimensional
#        self.gene2gkdb["dxdr"] = np.sqrt(np.interp(0.0,self.theta,self.geom.gxx,period=2.0*np.pi)) #CHECK THIS
        self.gene2gkdb["dxdr"] = 1.0/(0.5*(self.geom.dxdR[np.argmax(self.geom.R)]/self.geom.gxx[np.argmax(self.geom.R)]-\
                                      self.geom.dxdR[np.argmin(self.geom.R)]/self.geom.gxx[np.argmin(self.geom.R)]))
#        self.gene2gkdb["dxdr"] = 2.0/(self.geom.gxx[np.argmax(self.geom.R)]/self.geom.dxdR[np.argmax(self.geom.R)]-
#                                      self.geom.gxx[np.argmin(self.geom.R)]/self.geom.dxdR[np.argmin(self.geom.R)])

        self.gene2gkdb["Tref"] = 1.0/self.cm.pars["temp{}".format(ispec_electrons)] #Tref_GENE/Tref_GKDB
        self.gene2gkdb["mref"] = self.cm.pnt.mref/1.999007501778479 #mref_GENE/mref_GKDB (in units of proton mass)
        self.gene2gkdb["nref"] = 1.0/self.cm.pars["dens{}".format(ispec_electrons)] #nref_GENE/nref_GKDB
        self.gene2gkdb["qref"] = 1.0

        print ("Lref_GENE/Lref_GKDB = ", self.gene2gkdb["Lref"])
        print ("dx_GENE/dr_GKDB = ", self.gene2gkdb["dxdr"])

        #derived quantity conversion
        self.gene2gkdb["vth"]  = np.sqrt(0.5*self.gene2gkdb["Tref"]/self.gene2gkdb["mref"]) #cref_GENE/vth_GKDB
        self.gene2gkdb["rho"]  = self.gene2gkdb["mref"]*self.gene2gkdb["vth"]/(self.gene2gkdb["qref"]*self.gene2gkdb["Bref"])
        self.gene2gkdb["rho_Lref"] = self.gene2gkdb["rho"]/self.gene2gkdb["Lref"]

        print ("rho_Lref_GENE/rho_Lref_GKDB = ", self.gene2gkdb["rho_Lref"])

        #field conversion
        self.gene2gkdb["phi"] = self.gene2gkdb["Tref"]/self.gene2gkdb["qref"]*self.gene2gkdb["rho_Lref"]
        self.gene2gkdb["apar"] = self.gene2gkdb["Lref"]*self.gene2gkdb["Bref"]*self.gene2gkdb["rho_Lref"]**2
        self.gene2gkdb["bpar"] = self.gene2gkdb["Bref"]*self.gene2gkdb["rho_Lref"]

        #moments conversions (may be species dependent)
        self.gene2gkdb["dens"] = []
        self.gene2gkdb["upar"] = []
        self.gene2gkdb["tpar"] = []
        self.gene2gkdb["tperp"] = []
        for spec in self.cm.specnames:
            ispec = self.cm.specnames.index(spec) + 1
            self.gene2gkdb["dens"] += [self.gene2gkdb["nref"]*self.gene2gkdb["rho_Lref"]*\
                                       self.cm.pars["dens{}".format(ispec)]]
            self.gene2gkdb["upar"] += [self.gene2gkdb["vth"]*self.gene2gkdb["rho_Lref"]]
            self.gene2gkdb["tpar"] += [self.gene2gkdb["Tref"]*self.gene2gkdb["rho_Lref"]*\
                                       self.cm.pars["temp{}".format(ispec)]]
            self.gene2gkdb["tperp"] += [self.gene2gkdb["Tref"]*self.gene2gkdb["rho_Lref"]*\
                                        self.cm.pars["temp{}".format(ispec)]]

        #fluxes
        self.gene2gkdb["pflux"] = self.gene2gkdb["nref"]*self.gene2gkdb["vth"]*self.gene2gkdb["rho_Lref"]**2
        self.gene2gkdb["mflux"] = self.gene2gkdb["nref"]*self.gene2gkdb["mref"]*self.gene2gkdb["Lref"]*\
                                  (self.gene2gkdb["vth"]*self.gene2gkdb["rho_Lref"])**2   #LREF?
        self.gene2gkdb["qflux"] = self.gene2gkdb["Tref"]*self.gene2gkdb["pflux"]

    def fill_gkdb_from_par(self):
        """ Fill gkdb dict """
        pars = self.cm.pars
        geom = self.geom
        rat = self.gene2gkdb

        self.gkdict.clear()
        self.gkdict["ids_properties"] = {#"creator" : os.environ['USER'],
                                "creator" : self.creator,
                                "date" : str(datetime.date.today()),
                                "comment" : "json entry for testing only - note that "+\
                                "mode structure moments and flux-surface averaged fluxes "+\
                                "are given in particle space (i.e. incl. the full pullback operator)\n "+\
                                "EM flux contributions are for now *not* separated and all attributed "+\
                                "to the Apar term"
                            }
        self.gkdict["code"] = {"name": "GENE",
                               "version" : pars['RELEASE']+' - '+(pars['GIT_MASTER'])[0:9],
                               "parameters" : pars
                           }


        self.gkdict["model"] = {
            "initial_value_run": pars['comp_type']=="'IV'",
            "non_linear_run": self.cm.nonlinear,
            "time_interval_norm" : [], #FILL
            "include_a_field_parallel" : self.cm.electromagnetic,
            "include_b_field_parallel" : self.cm.bpar,
            "include_full_curvature_drift" : (True if pars['dpdx_pm']==0 else
                                              pars['dpdx_term']=='full_drift'),
            "include_centrifugal_effects" : pars['with_centrifugal'],
            "collision_pitch_only" : pars['collision_op']=="'pitch_angle'",
            "collision_ei_only" : False,
            "collision_momentum_conservation" : (pars['coll_cons_model']!='none'
                                                 if pars['coll'] > 0.0 else False),
            "collision_energy_conservation" :  (pars['coll_cons_model']!='none'
                                                if pars['coll'] > 0.0 else False),
            "collision_finite_larmor_radius" : pars['collision_op']=='sugama', #CHECK THIS (WITH_FLR = T ALSO?)
        }

        s_Ip = -pars['sign_Ip_CW']
        s_Bt = -pars['sign_Bt_CW']

        self.gkdict["flux_surface"] = {
            "r_minor_norm" : self.r/self.R0,
            "q" : s_Ip*s_Bt*np.abs(pars['q0']),
            "magnetic_shear_r_minor" : pars['shat']*self.r/rat["Lref"]/pars['x0']*rat["dxdr"], #CHECK Lref HERE
            "pressure_gradient_norm" : pars['dpdx_pm']/rat["Lref"]*rat["Bref"]**2*rat["dxdr"],
            "ip_sign" : s_Ip,
            "b_field_tor_sign" : s_Bt,
            # Original shape
            "shape_coefficients_c" : self.cN.tolist(), #NORMALIZATION?
            "shape_coefficients_s" : self.sN.tolist(), #NORMALIZATION?
            "dc_dr_minor_norm" : self.cN_dr.tolist(),
            "ds_dr_minor_norm" : self.sN_dr.tolist(),
            # Derived from Shape
            "elongation" : (pars['kappa'] if pars['magn_geometry']=="'miller'" else None),
            "triangularity_upper" : (pars['delta'] if pars['magn_geometry']=="'miller'" else None),
            "triangularity_lower" : (pars['delta'] if pars['magn_geometry']=="'miller'" else None),
        }


        self.gkdict["species_all"] = {
            "beta_reference" : pars['beta']/(rat["nref"]*rat["Tref"])*rat["Bref"]**2,
            "debye_length_reference" : np.sqrt(pars['debye2']), #ADD FACTORS
            "velocity_tor_norm" : pars["Omega0_tor"]*rat["vth"]/rat["Lref"],
            "shearing_rate_norm" : pars["ExBrate"]*self.r/pars["x0"]*\
                rat["vth"]/rat["Lref"]*rat["dxdr"],
            "zeff" : None #pars['Zeff'] #self-consistent computation by data base itself?
        }

        self.gkdict["species"] = []
        self.gkdict["collisions"] = {}
        self.gkdict["collisions"]["collisionality_norm"] = []

#        coulomb_log = 24.-np.log(np.sqrt(self.cm.pnt.nref*1E13)/
#                                self.cm.pnt.Tref*0.001)

        for ispec in range(1,pars['n_spec']+1):
            self.gkdict["species"] += [{
                "name" : self.cm.pars["name{}".format(ispec)],
                "charge_norm" : self.cm.pars["charge{}".format(ispec)],
                "mass_norm" : self.cm.pars["mass{}".format(ispec)]*rat["mref"],
                "density_norm" : (0.0 if self.cm.pars["passive{}".format(ispec)] else
                                  self.cm.pars["dens{}".format(ispec)]*rat["nref"]),
                "temperature_norm" : self.cm.pars["temp{}".format(ispec)]*rat["Tref"],
                "density_log_gradient_norm" : self.cm.pars["omn{}".format(ispec)]/rat["Lref"]*rat["dxdr"],
                "temperature_log_gradient_norm" : self.cm.pars["omt{}".format(ispec)]/rat["Lref"]*rat["dxdr"],
                "velocity_tor_gradient_norm" : pars["ExBrate"]*pars["q0"]/pars["x0"]*\
                rat["vth"]/rat["Lref"]*rat["dxdr"]
            }]

        for ispec in range(0,pars['n_spec']):
            ijcoll = []
            for jspec in range(0,pars['n_spec']):
                ijcoll += [(0.0 if self.cm.pars["passive{}".format(jspec+1)] else
                            rat["Tref"]**2/(np.sqrt(2.0)*np.pi*rat["qref"]**4*rat["Lref"]*rat["nref"])*\
                            self.gkdict["species"][ispec]["charge_norm"]**2*
                            self.gkdict["species"][jspec]["charge_norm"]**2*
                            self.gkdict["species"][jspec]["density_norm"]/\
                            np.sqrt(self.gkdict["species"][jspec]["mass_norm"]*
                                    self.gkdict["species"][jspec]["temperature_norm"]**3)*\
                            pars['coll'])]
            self.gkdict["collisions"]["collisionality_norm"] += [ijcoll]

        Cyq0_x0 = geom.Cy * self.cm.pnt.q0 / self.cm.pnt.x0
        if self.cm.pnt.adapt_lx:
            absnexc = 1
        else:
            absnexc = int(np.round(self.cm.pnt.lx*self.cm.pnt.n_pol*
                    np.abs(self.cm.pnt.shat)*self.cm.pnt.kymin*
                    np.abs(Cyq0_x0)))
            
#        nconn = int(int(int(self.cm.pnt.nx0-1)/2)/(absnexc*self.cm.pnt.ky0_ind))*2+1
        divisor = absnexc*self.cm.pnt.ky0_ind
        if divisor==0:
            nconn = 0
        else:
            nconn = int(int(int(self.cm.pnt.nx0-1)/2)/(absnexc*self.cm.pnt.ky0_ind))*2+1

        theta_full = []
        for iconn in range(-int(nconn/2),int(nconn/2)+1):
            theta_full += list(self.theta-np.sign(self.theta[0])*iconn*2.0*np.pi)
        fullsort_ind = np.argsort(theta_full)
        thetasort = np.array(theta_full)[fullsort_ind]


        #transformation of wavevector components on a z grid
        #following D. Told, PhD thesis, Sec. A.3, p. 167
        k1facx = np.sqrt(geom.gxx)
        k1facy = geom.gxy/np.sqrt(geom.gxx)
        k2fac = np.sqrt(geom.gyy-geom.gxy**2/geom.gxx)

        #now interpolate to theta = 0
        #interpolate to equidist. grid with proper order (interpol requires increasing arr)
        sort_ind = np.argsort(self.theta)
        k1facx_th0 = np.interp(0.0,self.theta[sort_ind],k1facx[sort_ind])/rat["rho"]
        k1facy_th0 = np.interp(0.0,self.theta[sort_ind],k1facy[sort_ind])/rat["rho"]
        k2fac_th0 = np.interp(0.0,self.theta[sort_ind],k2fac[sort_ind])/rat["rho"]

        ky = self.cm.pnt.kymin*self.cm.pnt.ky0_ind
        k1val = (self.cm.pnt.kx_center*k1facx_th0+ky*k1facy_th0)
        k2val = ky*k2fac_th0

        print ("k1facx_th0, k1facy_th0, k2fac_th0: ", k1facx_th0, k1facy_th0, k2fac_th0)

        # Finished writing physics inputs
        ###########################################################################################################
        # Now starting outputs

        eigendat = Eigenvaluedata(self.cm)

#        print('======{}========'.format(eigendat.growth_rate))
        if len(eigendat.growth_rate)>1 and max(eigendat.growth_rate) <= 0.0:
            raise RuntimeError("No unstable growth rate found - skipping file")

        nrgdat = self.rundatafiles.get_fileobject("nrg")
        nrgdat.generate_timeseries()

        newfieldnames = {"phi" : "phi_potential_perturbed_norm",
                         "apar" : "a_field_parallel_perturbed_norm",
                         "bpar" : "b_field_parallel_perturbed_norm"}

        newmomnames = { "dens" : "density",
                        "upar" : "velocity_parallel",
                        "tpar" : "temperature_parallel",
                        "tperp" : "temperature_perpendicular"}

        fieldmodestruct = ModeStructure(self.cm, self.rundatafiles,tavg=False)
        mommodestruct = []
        for spec in self.cm.specnames:
            ispec = self.cm.specnames.index(spec) #w/o +1
            mommodestruct += [ModeStructure(self.cm, self.rundatafiles, \
                                moms=list(newmomnames.keys()), species=spec,tavg=False)]

        eigenmodes = []

        if self.gkdict["model"]["initial_value_run"]:
            ievrange=[-1,]
        else:
            ievrange=range(len(eigendat.growth_rate))

        for iev in ievrange:
            if eigendat.growth_rate[iev] <= 0.0:
                continue

            eigenmode = {
                "growth_rate_norm" : eigendat.growth_rate[iev]*rat["vth"]/rat["Lref"],
                "frequency_norm" : eigendat.frequency[iev]*rat["vth"]/rat["Lref"],
                "growth_rate_tolerance": self.cm.pnt.omega_prec*rat["vth"]/rat["Lref"],
                "poloidal_angle" : thetasort.tolist()
            }

            #determine normalization amplitude and phase for parallel mode structures
            Af_gkdb = 0.0
            for field in self.fields:
                Af_gkdb += np.trapz(np.abs(fieldmodestruct.ballamps[field][iev][fullsort_ind]*rat[field])**2,thetasort)
            Af_gkdb = np.sqrt(Af_gkdb/(2.0*np.pi))
            alpha_gkdb = -np.interp(0.0,thetasort,np.angle(fieldmodestruct.ballamps["phi"][iev][fullsort_ind]))
            print ("Af_gkdb = ", Af_gkdb, ", alpha_gkdb = ", alpha_gkdb)

            for var in fieldmodestruct.ballamps.keys():
                print ("writing mode structure of ", var)
                eigenmode[newfieldnames[var]+"_real"]=np.real(fieldmodestruct.ballamps[var][iev][fullsort_ind]*
                                                        rat[var]*np.exp(1j*alpha_gkdb)/Af_gkdb).tolist()
                eigenmode[newfieldnames[var]+"_imaginary"]=np.imag(fieldmodestruct.ballamps[var][iev][fullsort_ind]*
                                                            rat[var]*np.exp(1j*alpha_gkdb)/Af_gkdb).tolist()

            eigenmode["moments_norm_rotating_frame"] = []
            for spec in self.cm.specnames:
                ispec = self.cm.specnames.index(spec) #w/o +1
#                mommodestruct = ModeStructure(self.cm, self.rundatafiles, moms=list(newmomnames.keys()), \
#                                                         species=spec,tavg=False)
                spec_moments_norm_rotating_frame = {}
                for var in newmomnames.keys():
                    spec_moments_norm_rotating_frame[newmomnames[var]+"_real"] = \
                                np.real(mommodestruct[ispec].ballamps[var][iev][fullsort_ind]*\
                                rat[var][ispec]*np.exp(1j*alpha_gkdb)/Af_gkdb).tolist()
                    spec_moments_norm_rotating_frame[newmomnames[var]+"_imaginary"] = \
                                np.imag(mommodestruct[ispec].ballamps[var][iev][fullsort_ind]*\
                                rat[var][ispec]*np.exp(1j*alpha_gkdb)/Af_gkdb).tolist()
                eigenmode["moments_norm_rotating_frame"] += [spec_moments_norm_rotating_frame]


            eigenmode["fluxes_norm"] = []
            #NOTE: GENE NRG-FILES DO NOT ALLOW FOR SEPARATION OF A/B_FIELD_PARALLEL CONTRIBUTIONS
            #CURRENTLY BOTH OF THEM ARE ATTRIBUTED TO A_FIELD_PARALLEL
            #ALTERNATIVE: RECOMPUTE FLUXES FROM MOM-FILES (Pullback will still contain both fields though)
            for spec in self.cm.specnames:
                ispec = self.cm.specnames.index(spec) #w/o +1
                eigenmode["fluxes_norm"] += [{
                    "particles_phi_potential": nrgdat.dataarray[iev,ispec,4]*rat["pflux"]*(1.0/Af_gkdb)**2,
                    "particles_a_field_parallel": nrgdat.dataarray[iev,ispec,5]*rat["pflux"]*(rat["rho_Lref"]/Af_gkdb)**2,      #FIX THIS
                    "particles_b_field_parallel": nrgdat.dataarray[iev,ispec,5]*rat["pflux"]*(1.0/Af_gkdb)**2*0.0,  #FIX THIS -- SET TO ZERO
                    "momentum_tor_parallel_phi_potential": nrgdat.dataarray[iev,ispec,8]*rat["mflux"]*(1.0/Af_gkdb)**2,
                    "momentum_tor_parallel_a_field_parallel": nrgdat.dataarray[iev,ispec,9]*rat["mflux"]*(1.0/Af_gkdb)**2,      #FIX THIS
                    "momentum_tor_parallel_b_field_parallel": nrgdat.dataarray[iev,ispec,9]*rat["mflux"]*(1.0/Af_gkdb)**2*0.0,  #FIX THIS -- SET TO ZERO
                    "momentum_tor_perpendicular_phi_potential": None,
                    "momentum_tor_perpendicular_a_field_parallel": None,
                    "momentum_tor_perpendicular_b_field_parallel": None,
                    "energy_phi_potential": nrgdat.dataarray[iev,ispec,6]*rat["qflux"]*(1.0/Af_gkdb)**2,
                    "energy_a_field_parallel": nrgdat.dataarray[iev,ispec,7]*rat["qflux"]*(1.0/Af_gkdb)**2,     #FIX THIS
                    "energy_b_field_parallel": nrgdat.dataarray[iev,ispec,7]*rat["qflux"]*(1.0/Af_gkdb)**2*0.0  #FIX THIS -- SET TO ZERO
                }]
            eigenmodes += [eigenmode]


        self.gkdict["wavevector"] = [{
            "radial_component_norm" : k1val,
            "binormal_component_norm" : k2val,
            "poloidal_turns" : nconn,
            "eigenmode" : eigenmodes
        }]

        self.gkdict["fluxes_integrated_norm"] = []

    def Write_json(self, path):
        """ Take the dict and write a gkdb json parameters file """
        #        print(json.dumps(self.gkdict, indent=4))
        filename = 'data{}.json'.format(self.cm.fileextension)
        with open(filename, 'w') as fp:
            json.dump(self.gkdict, fp, indent=4,separators=(',', ': '))



class GKDB_nonlin(object):
    """ GKDB class:

    Converts the results of a nonlinear GENE run into a GKDB-style json file

    """

    def __init__(self,common,rundatafiles,creator):
        # dictionary for gkdb parameters: {parameter: value}
        self.creator = creator
        self.cm = common
#        self.precheck()
        self.geom = Geometry(self.cm)
        self.fields = ["phi"]
        if common.electromagnetic:
            self.fields += ["apar"]
            if common.bpar:
                self.fields += ["bpar"]

        self.rundatafiles = rundatafiles
        self.gkdict = OrderedDict()
        self.set_gkdb_shape()
        self.set_gene2gkdb_conversion()
        self.fill_gkdb_from_par()
        

    def precheck(self):
        """ Run some basic checks before actual execution """
        if self.cm.nonlinear:
            raise RuntimeError("Nonlinear simulations not yet supported")
        else:
            if self.cm.pars['comp_type']=="'IV'":
                if self.cm.starttime!=-2 or self.cm.endtime!=-2:
                    print("*** WARNING: Selecting l:l as time window could be more appropriate ***")
            else:
                if (self.cm.starttime!=-1 and self.cm.starttime!=1) or \
                   (self.cm.endtime!=-2 and self.cm.endtime!=self.cm.pars["n_ev"]):
                    raise RuntimeError("Select f:l as time window for EV simulations")
        if self.cm.pars['magn_geometry']!="'miller'":
            raise RuntimeError("Only Miller geometries supported for now")


    def set_gkdb_shape(self):
        """ Determines shape parameters for GKDB """

        print ("Determining flux surface shape ...")
        nz0 = self.cm.pnt.nz0

        #determine flux surface center
        self.R0 = 0.5*(max(self.geom.R)+min(self.geom.R))
        self.Z0 = 0.5*(max(self.geom.Z)+min(self.geom.Z))
        print ("(R0,Z0) = (", self.R0, ",", self.Z0,")")
#        self.R0 = np.sum(self.geom.R*self.geom.jacobian)/np.sum(self.geom.jacobian)
#        self.Z0 = np.sum(self.geom.Z*self.geom.jacobian)/np.sum(self.geom.jacobian)

        #minor radius of flux surface
        self.r = 0.5*(max(self.geom.R)-min(self.geom.R))

        #poloidal theta (COCOS=11)
        self.theta = np.arctan(-(self.geom.Z-self.Z0)/(self.geom.R-self.R0))
        for iz in range(int(nz0/2),nz0-1):
            thetadiff = (self.theta[iz+1]-self.theta[iz])
            if thetadiff>0.5*np.pi:
                self.theta[iz+1]-=np.pi
            elif thetadiff<-0.5*np.pi:
                self.theta[iz+1]+=np.pi
            thetadiff = (self.theta[nz0-1-iz-1]-self.theta[nz0-1-iz])
            if thetadiff>0.5*np.pi:
                self.theta[nz0-1-iz-1]-=np.pi
            elif thetadiff<-0.5*np.pi:
                self.theta[nz0-1-iz-1]+=np.pi

        #sorted theta (starting at th=0)
        theta_sort = (self.theta+2.0*np.pi)%(2.0*np.pi)
        sort_ind = np.argsort(theta_sort)
        theta_sort = theta_sort[sort_ind]

        #Flux surface distance (original grid and normalized)
        aN = np.sqrt((self.geom.R[sort_ind]-self.R0)**2+(self.geom.Z[sort_ind]-self.Z0)**2)/self.R0
        aN_dr = ((self.geom.R[sort_ind]-self.R0)*self.geom.dxdR[sort_ind]+
                 (self.geom.Z[sort_ind]-self.Z0)*self.geom.dxdZ[sort_ind])/\
            (aN*self.geom.gxx[sort_ind])/self.R0

        #equidistant theta (COCOS=11) with somewhat higher resolution for fourier transforms etc
        nth = 4*nz0
        theta_grid = (np.arange(0.0,nth,1.0))*2.0*np.pi/nth

        #interpolate to this equidist. grid with proper order (interpol requires increasing arr)
        aN_th = np.interp(theta_grid,theta_sort,aN,period=2.0*np.pi)
        aN_dr_th = np.interp(theta_grid,theta_sort,aN_dr,period=2.0*np.pi)

        #now fourier transform
        fft_aN = np.fft.fft(aN_th)
        fft_aN_dr = np.fft.fft(aN_dr_th)

        #keep up to nth_kept modes (typically around 8)
        nth_kept = 8 #nz0
        self.cN = 2.0*np.real(fft_aN)[0:nth_kept]/nth
        self.cN[0] *= 0.5
        self.sN = -2.0*np.imag(fft_aN)[0:nth_kept]/nth
        self.sN[0] *= 0.5
        self.cN_dr = 2.0*np.real(fft_aN_dr)[0:nth_kept]/nth
        self.cN_dr[0] *= 0.5
        self.sN_dr = -2.0*np.imag(fft_aN_dr)[0:nth_kept]/nth
        self.sN_dr[0] *= 0.5

        #check parametrization
        nind = np.arange(0.0,nth_kept,1.0)
        a_check = []
        a_dr_check = []
        for iz in range(0,nth):
            a_check += [np.sum(self.cN*np.cos(nind*theta_grid[iz])+
                               self.sN*np.sin(nind*theta_grid[iz]))]
            a_dr_check += [np.sum(self.cN_dr*np.cos(nind*theta_grid[iz])+
                                  self.sN_dr*np.sin(nind*theta_grid[iz]))]

        a_dr_check = np.array(a_dr_check)
        err = np.sum(np.abs(a_check-aN_th)/aN_th)/nth
        print ("Relative error of flux surface parametrization: {0:12.6E}".format(err))

        #some plot diagnostics
        show_aN = False
        if show_aN:
            plt.plot(theta_sort,aN,label='aN_orig')
            plt.plot(theta_grid,aN_th,label='aN_th,intermediate')
            plt.plot(theta_grid,a_check,label='aN_final')
            plt.plot(theta_sort,aN_dr,label='aN_dr_orig')
            plt.plot(theta_grid,a_dr_check,label='aN_dr_final')
            plt.legend()
            plt.show()

        show_flux_surface = False
        if show_flux_surface:
            dr = 0.04  #radial extension for testing

            plt.plot(self.geom.R[sort_ind],self.geom.Z[sort_ind],'o',label="original")
            plt.plot((self.geom.R[sort_ind]+dr*self.geom.dxdR[sort_ind]/self.geom.gxx[sort_ind]),
                     (self.geom.Z[sort_ind]+dr*self.geom.dxdZ[sort_ind]/self.geom.gxx[sort_ind]),
                     'x',label="original+dr")
            theta_grid = (np.arange(0.0,nth,1.0))*2.0*np.pi/(nth-1)
            # R_n_th = self.R0+aN_th*np.cos(theta_grid)
            # Z_n_th = self.Z0-aN_th*np.sin(theta_grid)
            # plt.plot(R_n_th,Z_n_th,label="intermediate")
            R_check = self.R0+a_check*np.cos(theta_grid)*self.R0
            Z_check = self.Z0-a_check*np.sin(theta_grid)*self.R0
            plt.plot(R_check,Z_check,label="gkdb")
            R_check = self.R0+(a_check+dr/self.R0*a_dr_check)*np.cos(theta_grid)*self.R0
            Z_check = self.Z0-(a_check+dr/self.R0*a_dr_check)*np.sin(theta_grid)*self.R0
            plt.plot(R_check,Z_check,label="gkdb+dr")
            plt.xlabel('R/m')
            plt.ylabel('Z/m')
            plt.axis('equal')
            plt.legend()
            plt.show()

    def set_gene2gkdb_conversion(self):
        """ set conversion factors for various quantities """

        print ("Setting conversion factors ...")

        #determine index of electron species
        ispec_electrons = -1111
        for spec in self.cm.specnames:
            ispec = self.cm.specnames.index(spec) + 1
            if self.cm.pars["charge{}".format(ispec)]==-1:
                ispec_electrons = ispec

        self.gene2gkdb = OrderedDict()
        self.gene2gkdb["Lref"] = self.cm.pnt.Lref/self.R0 #Lref_GENE/Lref_GKDB
        self.gene2gkdb["Bref"] = self.cm.pnt.Lref*self.cm.pnt.major_R/self.R0  #JUST 1/R correction for the moment
        print ("Bref_GENE/Bref_GKDB = ", self.gene2gkdb["Bref"])

        #CHECK THIS: x dimensionless, r dimensional
#        self.gene2gkdb["dxdr"] = np.sqrt(np.interp(0.0,self.theta,self.geom.gxx,period=2.0*np.pi)) #CHECK THIS
        self.gene2gkdb["dxdr"] = 1.0/(0.5*(self.geom.dxdR[np.argmax(self.geom.R)]/self.geom.gxx[np.argmax(self.geom.R)]-\
                                      self.geom.dxdR[np.argmin(self.geom.R)]/self.geom.gxx[np.argmin(self.geom.R)]))
#        self.gene2gkdb["dxdr"] = 2.0/(self.geom.gxx[np.argmax(self.geom.R)]/self.geom.dxdR[np.argmax(self.geom.R)]-
#                                      self.geom.gxx[np.argmin(self.geom.R)]/self.geom.dxdR[np.argmin(self.geom.R)])

        self.gene2gkdb["Tref"] = 1.0/self.cm.pars["temp{}".format(ispec_electrons)] #Tref_GENE/Tref_GKDB
        self.gene2gkdb["mref"] = self.cm.pnt.mref/1.999007501778479 #mref_GENE/mref_GKDB (in units of proton mass)
        self.gene2gkdb["nref"] = 1.0/self.cm.pars["dens{}".format(ispec_electrons)] #nref_GENE/nref_GKDB
        self.gene2gkdb["qref"] = 1.0

        print ("Lref_GENE/Lref_GKDB = ", self.gene2gkdb["Lref"])
        print ("dx_GENE/dr_GKDB = ", self.gene2gkdb["dxdr"])

        #derived quantity conversion
        self.gene2gkdb["vth"]  = np.sqrt(0.5*self.gene2gkdb["Tref"]/self.gene2gkdb["mref"]) #cref_GENE/vth_GKDB
        self.gene2gkdb["rho"]  = self.gene2gkdb["mref"]*self.gene2gkdb["vth"]/(self.gene2gkdb["qref"]*self.gene2gkdb["Bref"])
        self.gene2gkdb["rho_Lref"] = self.gene2gkdb["rho"]/self.gene2gkdb["Lref"]

        print ("rho_Lref_GENE/rho_Lref_GKDB = ", self.gene2gkdb["rho_Lref"])

        #field conversion
        self.gene2gkdb["phi"] = self.gene2gkdb["Tref"]/self.gene2gkdb["qref"]*self.gene2gkdb["rho_Lref"]
        self.gene2gkdb["apar"] = self.gene2gkdb["Lref"]*self.gene2gkdb["Bref"]*self.gene2gkdb["rho_Lref"]**2
        self.gene2gkdb["bpar"] = self.gene2gkdb["Bref"]*self.gene2gkdb["rho_Lref"]

        #moments conversions (may be species dependent)
        self.gene2gkdb["dens"] = []
        self.gene2gkdb["upar"] = []
        self.gene2gkdb["tpar"] = []
        self.gene2gkdb["tperp"] = []
        for spec in self.cm.specnames:
            ispec = self.cm.specnames.index(spec) + 1
            self.gene2gkdb["dens"] += [self.gene2gkdb["nref"]*self.gene2gkdb["rho_Lref"]*\
                                       self.cm.pars["dens{}".format(ispec)]]
            self.gene2gkdb["upar"] += [self.gene2gkdb["vth"]*self.gene2gkdb["rho_Lref"]]
            self.gene2gkdb["tpar"] += [self.gene2gkdb["Tref"]*self.gene2gkdb["rho_Lref"]*\
                                       self.cm.pars["temp{}".format(ispec)]]
            self.gene2gkdb["tperp"] += [self.gene2gkdb["Tref"]*self.gene2gkdb["rho_Lref"]*\
                                        self.cm.pars["temp{}".format(ispec)]]

        #fluxes
        self.gene2gkdb["pflux"] = self.gene2gkdb["nref"]*self.gene2gkdb["vth"]*self.gene2gkdb["rho_Lref"]**2
        self.gene2gkdb["mflux"] = self.gene2gkdb["nref"]*self.gene2gkdb["mref"]*self.gene2gkdb["Lref"]*\
                                  (self.gene2gkdb["vth"]*self.gene2gkdb["rho_Lref"])**2   #LREF?
        self.gene2gkdb["qflux"] = self.gene2gkdb["Tref"]*self.gene2gkdb["pflux"]

    def fill_gkdb_from_par(self):
        """ Fill gkdb dict """
        pars = self.cm.pars
        geom = self.geom
        rat = self.gene2gkdb

        self.gkdict.clear()
        self.gkdict["ids_properties"] = {#"creator" : os.environ['USER'],
                                "creator" : self.creator,
                                "date" : str(datetime.date.today()),
                                "comment" : "json entry for testing only - note that "+\
                                "mode structure moments and flux-surface averaged fluxes "+\
                                "are given in particle space (i.e. incl. the full pullback operator)\n "+\
                                "EM flux contributions are for now *not* separated and all attributed "+\
                                "to the Apar term"
                            }
        self.gkdict["code"] = {"name": "GENE",
                               "version" : pars['RELEASE']+' - '+(pars['GIT_MASTER'])[0:9],
                               "parameters" : pars
                           }


        self.gkdict["model"] = {
            "initial_value_run": pars['comp_type']=="'IV'",
            "non_linear_run": self.cm.nonlinear,
            "time_interval_norm" : [], #FILL
            "include_a_field_parallel" : self.cm.electromagnetic,
            "include_b_field_parallel" : self.cm.bpar,
            "include_full_curvature_drift" : (True if pars['dpdx_pm']==0 else
                                              pars['dpdx_term']=='full_drift'),
            "include_centrifugal_effects" : pars['with_centrifugal'],
            "collision_pitch_only" : pars['collision_op']=="'pitch_angle'",
            "collision_ei_only" : False,
            "collision_momentum_conservation" : (pars['coll_cons_model']!='none'
                                                 if pars['coll'] > 0.0 else False),
            "collision_energy_conservation" :  (pars['coll_cons_model']!='none'
                                                if pars['coll'] > 0.0 else False),
            "collision_finite_larmor_radius" : pars['collision_op']=='sugama', #CHECK THIS (WITH_FLR = T ALSO?)
        }

        s_Ip = -pars['sign_Ip_CW']
        s_Bt = -pars['sign_Bt_CW']

        self.gkdict["flux_surface"] = {
            "r_minor_norm" : self.r/self.R0,
            "q" : s_Ip*s_Bt*np.abs(pars['q0']),
            "magnetic_shear_r_minor" : pars['shat']*self.r/rat["Lref"]/pars['x0']*rat["dxdr"], #CHECK Lref HERE
            "pressure_gradient_norm" : pars['dpdx_pm']/rat["Lref"]*rat["Bref"]**2*rat["dxdr"],
            "ip_sign" : s_Ip,
            "b_field_tor_sign" : s_Bt,
            # Original shape
            "shape_coefficients_c" : self.cN.tolist(), #NORMALIZATION?
            "shape_coefficients_s" : self.sN.tolist(), #NORMALIZATION?
            "dc_dr_minor_norm" : self.cN_dr.tolist(),
            "ds_dr_minor_norm" : self.sN_dr.tolist(),
            # Derived from Shape
            "elongation" : (pars['kappa'] if pars['magn_geometry']=="'miller'" else None),
            "triangularity_upper" : (pars['delta'] if pars['magn_geometry']=="'miller'" else None),
            "triangularity_lower" : (pars['delta'] if pars['magn_geometry']=="'miller'" else None),
        }


        self.gkdict["species_all"] = {
            "beta_reference" : pars['beta']/(rat["nref"]*rat["Tref"])*rat["Bref"]**2,
            "debye_length_reference" : np.sqrt(pars['debye2']), #ADD FACTORS
            "velocity_tor_norm" : pars["Omega0_tor"]*rat["vth"]/rat["Lref"],
            "shearing_rate_norm" : pars["ExBrate"]*self.r/pars["x0"]*\
                rat["vth"]/rat["Lref"]*rat["dxdr"],
            "zeff" : None #pars['Zeff'] #self-consistent computation by data base itself?
        }

        self.gkdict["species"] = []
        self.gkdict["collisions"] = {}
        self.gkdict["collisions"]["collisionality_norm"] = []

#        coulomb_log = 24.-np.log(np.sqrt(self.cm.pnt.nref*1E13)/
#                                self.cm.pnt.Tref*0.001)

        for ispec in range(1,pars['n_spec']+1):
            self.gkdict["species"] += [{
                "name" : self.cm.pars["name{}".format(ispec)],
                "charge_norm" : self.cm.pars["charge{}".format(ispec)],
                "mass_norm" : self.cm.pars["mass{}".format(ispec)]*rat["mref"],
                "density_norm" : (0.0 if self.cm.pars["passive{}".format(ispec)] else
                                  self.cm.pars["dens{}".format(ispec)]*rat["nref"]),
                "temperature_norm" : self.cm.pars["temp{}".format(ispec)]*rat["Tref"],
                "density_log_gradient_norm" : self.cm.pars["omn{}".format(ispec)]/rat["Lref"]*rat["dxdr"],
                "temperature_log_gradient_norm" : self.cm.pars["omt{}".format(ispec)]/rat["Lref"]*rat["dxdr"],
                "velocity_tor_gradient_norm" : pars["ExBrate"]*pars["q0"]/pars["x0"]*\
                rat["vth"]/rat["Lref"]*rat["dxdr"]
            }]

        for ispec in range(0,pars['n_spec']):
            ijcoll = []
            for jspec in range(0,pars['n_spec']):
                ijcoll += [(0.0 if self.cm.pars["passive{}".format(jspec+1)] else
                            rat["Tref"]**2/(np.sqrt(2.0)*np.pi*rat["qref"]**4*rat["Lref"]*rat["nref"])*\
                            self.gkdict["species"][ispec]["charge_norm"]**2*
                            self.gkdict["species"][jspec]["charge_norm"]**2*
                            self.gkdict["species"][jspec]["density_norm"]/\
                            np.sqrt(self.gkdict["species"][jspec]["mass_norm"]*
                                    self.gkdict["species"][jspec]["temperature_norm"]**3)*\
                            pars['coll'])]
            self.gkdict["collisions"]["collisionality_norm"] += [ijcoll]

        Cyq0_x0 = geom.Cy * self.cm.pnt.q0 / self.cm.pnt.x0
        if self.cm.pnt.adapt_lx:
            absnexc = 1
        else:
            absnexc = int(np.round(self.cm.pnt.lx*self.cm.pnt.n_pol*
                    np.abs(self.cm.pnt.shat)*self.cm.pnt.kymin*
                    np.abs(Cyq0_x0)))
            
#        nconn = int(int(int(self.cm.pnt.nx0-1)/2)/(absnexc*self.cm.pnt.ky0_ind))*2+1
        divisor = absnexc*self.cm.pnt.ky0_ind
        if divisor==0:
            nconn = 0
        else:
            nconn = int(int(int(self.cm.pnt.nx0-1)/2)/(absnexc*self.cm.pnt.ky0_ind))*2+1

        theta_full = []
        for iconn in range(-int(nconn/2),int(nconn/2)+1):
            theta_full += list(self.theta-np.sign(self.theta[0])*iconn*2.0*np.pi)
        fullsort_ind = np.argsort(theta_full)
        thetasort = np.array(theta_full)[fullsort_ind]


        #transformation of wavevector components on a z grid
        #following D. Told, PhD thesis, Sec. A.3, p. 167
        k1facx = np.sqrt(geom.gxx)
        k1facy = geom.gxy/np.sqrt(geom.gxx)
        k2fac = np.sqrt(geom.gyy-geom.gxy**2/geom.gxx)

        #now interpolate to theta = 0
        #interpolate to equidist. grid with proper order (interpol requires increasing arr)
        sort_ind = np.argsort(self.theta)
        k1facx_th0 = np.interp(0.0,self.theta[sort_ind],k1facx[sort_ind])/rat["rho"]
        k1facy_th0 = np.interp(0.0,self.theta[sort_ind],k1facy[sort_ind])/rat["rho"]
        k2fac_th0 = np.interp(0.0,self.theta[sort_ind],k2fac[sort_ind])/rat["rho"]

        ky = self.cm.pnt.kymin*self.cm.pnt.ky0_ind
        k1val = (self.cm.pnt.kx_center*k1facx_th0+ky*k1facy_th0)
        k2val = ky*k2fac_th0

        print ("k1facx_th0, k1facy_th0, k2fac_th0: ", k1facx_th0, k1facy_th0, k2fac_th0)

        # Finished writing physics inputs
        ###########################################################################################################
        # Now starting outputs

#        eigendat = Eigenvaluedata(self.cm)
#
#        if max(eigendat.growth_rate) <= 0.0:
#            raise RuntimeError("No unstable growth rate found - skipping file")

        nrgdat = self.rundatafiles.get_fileobject("nrg")
        nrgdat.generate_timeseries()

        newfieldnames = {"phi" : "phi_potential_perturbed_norm",
                         "apar" : "a_field_parallel_perturbed_norm",
                         "bpar" : "b_field_parallel_perturbed_norm"}

        newmomnames = { "dens" : "density",
                        "upar" : "velocity_parallel",
                        "tpar" : "temperature_parallel",
                        "tperp" : "temperature_perpendicular"}

        fieldmodestruct = ModeStructure(self.cm, self.rundatafiles,tavg=False)
        mommodestruct = []
        for spec in self.cm.specnames:
            ispec = self.cm.specnames.index(spec) #w/o +1
            mommodestruct += [ModeStructure(self.cm, self.rundatafiles, \
                                moms=list(newmomnames.keys()), species=spec,tavg=False)]

        eigenmodes = []

        if self.gkdict["model"]["initial_value_run"]:
            ievrange=[-1,]
        else:
#            ievrange=range(len(eigendat.growth_rate))
            ievrange=range(1)

        for iev in ievrange:
#            if eigendat.growth_rate[iev] <= 0.0:
#                continue

            eigenmode = {
#                "growth_rate_norm" : eigendat.growth_rate[iev]*rat["vth"]/rat["Lref"],
#                "frequency_norm" : eigendat.frequency[iev]*rat["vth"]/rat["Lref"],
                "growth_rate_tolerance": self.cm.pnt.omega_prec*rat["vth"]/rat["Lref"],
                "poloidal_angle" : thetasort.tolist()
            }

            #determine normalization amplitude and phase for parallel mode structures
            Af_gkdb = 0.0
            for field in self.fields:
                Af_gkdb += np.trapz(np.abs(fieldmodestruct.ballamps[field][iev][fullsort_ind]*rat[field])**2,thetasort)
            Af_gkdb = np.sqrt(Af_gkdb/(2.0*np.pi))
            alpha_gkdb = -np.interp(0.0,thetasort,np.angle(fieldmodestruct.ballamps["phi"][iev][fullsort_ind]))
            print ("Af_gkdb = ", Af_gkdb, ", alpha_gkdb = ", alpha_gkdb)

            for var in fieldmodestruct.ballamps.keys():
                print ("writing mode structure of ", var)
                eigenmode[newfieldnames[var]+"_real"]=np.real(fieldmodestruct.ballamps[var][iev][fullsort_ind]*
                                                        rat[var]*np.exp(1j*alpha_gkdb)/Af_gkdb).tolist()
                eigenmode[newfieldnames[var]+"_imaginary"]=np.imag(fieldmodestruct.ballamps[var][iev][fullsort_ind]*
                                                            rat[var]*np.exp(1j*alpha_gkdb)/Af_gkdb).tolist()

            eigenmode["moments_norm_rotating_frame"] = []
            for spec in self.cm.specnames:
                ispec = self.cm.specnames.index(spec) #w/o +1
#                mommodestruct = ModeStructure(self.cm, self.rundatafiles, moms=list(newmomnames.keys()), \
#                                                         species=spec,tavg=False)
                spec_moments_norm_rotating_frame = {}
                for var in newmomnames.keys():
                    spec_moments_norm_rotating_frame[newmomnames[var]+"_real"] = \
                                np.real(mommodestruct[ispec].ballamps[var][iev][fullsort_ind]*\
                                rat[var][ispec]*np.exp(1j*alpha_gkdb)/Af_gkdb).tolist()
                    spec_moments_norm_rotating_frame[newmomnames[var]+"_imaginary"] = \
                                np.imag(mommodestruct[ispec].ballamps[var][iev][fullsort_ind]*\
                                rat[var][ispec]*np.exp(1j*alpha_gkdb)/Af_gkdb).tolist()
                eigenmode["moments_norm_rotating_frame"] += [spec_moments_norm_rotating_frame]


            eigenmode["fluxes_norm"] = []
            #NOTE: GENE NRG-FILES DO NOT ALLOW FOR SEPARATION OF A/B_FIELD_PARALLEL CONTRIBUTIONS
            #CURRENTLY BOTH OF THEM ARE ATTRIBUTED TO A_FIELD_PARALLEL
            #ALTERNATIVE: RECOMPUTE FLUXES FROM MOM-FILES (Pullback will still contain both fields though)
            for spec in self.cm.specnames:
                ispec = self.cm.specnames.index(spec) #w/o +1
                eigenmode["fluxes_norm"] += [{
                    "particles_phi_potential": nrgdat.dataarray[iev,ispec,4]*rat["pflux"]*(1.0/Af_gkdb)**2,
                    "particles_a_field_parallel": nrgdat.dataarray[iev,ispec,5]*rat["pflux"]*(rat["rho_Lref"]/Af_gkdb)**2,      #FIX THIS
                    "particles_b_field_parallel": nrgdat.dataarray[iev,ispec,5]*rat["pflux"]*(1.0/Af_gkdb)**2*0.0,  #FIX THIS -- SET TO ZERO
                    "momentum_tor_parallel_phi_potential": nrgdat.dataarray[iev,ispec,8]*rat["mflux"]*(1.0/Af_gkdb)**2,
                    "momentum_tor_parallel_a_field_parallel": nrgdat.dataarray[iev,ispec,9]*rat["mflux"]*(1.0/Af_gkdb)**2,      #FIX THIS
                    "momentum_tor_parallel_b_field_parallel": nrgdat.dataarray[iev,ispec,9]*rat["mflux"]*(1.0/Af_gkdb)**2*0.0,  #FIX THIS -- SET TO ZERO
                    "momentum_tor_perpendicular_phi_potential": None,
                    "momentum_tor_perpendicular_a_field_parallel": None,
                    "momentum_tor_perpendicular_b_field_parallel": None,
                    "energy_phi_potential": nrgdat.dataarray[iev,ispec,6]*rat["qflux"]*(1.0/Af_gkdb)**2,
                    "energy_a_field_parallel": nrgdat.dataarray[iev,ispec,7]*rat["qflux"]*(1.0/Af_gkdb)**2,     #FIX THIS
                    "energy_b_field_parallel": nrgdat.dataarray[iev,ispec,7]*rat["qflux"]*(1.0/Af_gkdb)**2*0.0  #FIX THIS -- SET TO ZERO
                }]
            eigenmodes += [eigenmode]


        self.gkdict["wavevector"] = [{
            "radial_component_norm" : k1val,
            "binormal_component_norm" : k2val,
            "poloidal_turns" : nconn,
            "eigenmode" : eigenmodes
        }]

        self.gkdict["fluxes_integrated_norm"] = []




    def Write_json(self, path):
        """ Take the dict and write a gkdb json parameters file """
        #        print(json.dumps(self.gkdict, indent=4))
        filename = 'data{}.json'.format(self.cm.fileextension)
        with open(filename, 'w') as fp:
            json.dump(self.gkdict, fp, indent=4,separators=(',', ': '))
            #json.dump(self.gkdict, fp, sort_keys=True,indent=4,separators=(',', ': '))



# -*- coding: utf-8 -*-

from tkinter import END

import h5py
import matplotlib.pyplot as plt
import numpy as np
import scipy.linalg as linalg

import ..utils.aux_func as aux_func
from .baseplot import Plotting
from .diagnostic import Diagnostic
import os

class DiagFrequencyGrowthRate(Diagnostic):
    # pylint: disable=invalid-name
    def __init__(self, avail_vars=None, specnames=None):
        super().__init__()
        self.name = 'Frequency & Growth-rate'
        self.tabs = ['fluxtube', 'xglobal', 'GENE3D']

        self.help_txt = """Frequency diagnostic
                        \n
                        \nWeight: weight with parallel mode structure (def. False)
                        \nky ind : comma separated |ky| fourier modes (e.g 0.5,1.0), default all range
                        \nz ind : comma separated z interval (e.g., -3,3), default all range
                        \nshow time traces : show time traces for different |ky|, default false
                        \nSave h5 : save hdf5 file (def. True)"""

        self.avail_vars = avail_vars
        self.specnames = specnames

        self.opts = {"weight": {'tag': 'weight', 'values': [False, True]},
                     "kyind": {'tag': "ky ind", 'values': None},
                     "zind": {'tag': "z ind", 'values': None},
                     "show": {'tag': "Show time traces", 'values': [False, True]}}

        self.set_defaults()

        self.options_gui = Diagnostic.OptionsGUI()

    def set_options(self, run, specnames, out_folder):

        self.phi = []
#        self.geom = run_data.geometry
#        self.run_data = run_data
        self.specnames = specnames
        
        self.pnt = run.parameters[0].asnamedtuple()

        self.weight = 1 if self.opts['weight']['value'] else 0
        
        self.ky = run.spatialgrid[0].ky

        kyind = self.opts['kyind']['value']
        if not kyind or kyind == -1:
            self.kyind = 0.0, 100
        else:
            self.kyind = kyind.split(',')
        self.kyind_min = aux_func.find_nearest(run.spatialgrid[0].ky, float(self.kyind[0]))
        self.kyind_max = aux_func.find_nearest(run.spatialgrid[0].ky, float(self.kyind[1]))

        # z position, unset or -1 means all
        zind = self.opts['zind']['value']
        if not zind or zind == -1:
            self.zind = -100, 100  # this will choose the minimum and maximum values
        else:
            self.zind = zind.split(',')
        self.zind_min, self.zind_max = aux_func.find_nearest_points(run.spatialgrid[0].z, float(self.zind[0]), float(self.zind[1]))

        self.get_needed_vars(['field'])

        return self.needed_vars

    def execute(self, data, parameters, geometry, spatialgrid,  step):

        phi = getattr(getattr(data, 'field'), "phi")(step.time, step.field, parameters)
        renorm = 0.0

        if self.pnt.flux_tube:
            phi_ = np.zeros(len(range(self.kyind_min, self.kyind_max+1)), dtype=np.complex64)
            # loop over all ky modes
            for i_ky, j in enumerate(range(self.kyind_min, self.kyind_max+1)):
                for i in range(0, self.pnt.nx0-1):
                    renorm = renorm + np.dot(np.squeeze(phi[i, j, self.zind_min:self.zind_max+1]**(2 * self.weight)),
                                             geometry.jacobian[self.zind_min:self.zind_max+1])
                    phi_[i_ky] = phi_[i_ky] + \
                        np.dot(np.squeeze(phi[i, j, self.zind_min:self.zind_max+1]**(2 * self.weight) * phi[i, j, self.zind_min:self.zind_max+1]),
                               geometry.jacobian[self.zind_min:self.zind_max+1])
                phi_[i_ky] = phi_[i_ky] / renorm

        elif self.pnt.x_global:
            phi_ = np.zeros(len(range(self.kyind_min, self.kyind_max+1)), dtype=np.complex64)
            # loop over all ky modes
            for i_ky, j in enumerate(range(self.kyind_min, self.kyind_max+1)):
                for i in range(0, self.pnt.nx0-1):
                    renorm = renorm + np.dot(np.squeeze(phi[i, i_ky, self.zind_min:self.zind_max+1]**(2 * self.weight)),
                                             geometry.jacobian.T[i, self.zind_min:self.zind_max+1])
                    phi_[i_ky] = phi_[i_ky] + \
                        np.dot(np.squeeze(phi[i, i_ky, self.zind_min:self.zind_max+1]**(2 * self.weight) * phi[i, i_ky, self.zind_min:self.zind_max+1]),
                               geometry.jacobian.T[i, self.zind_min:self.zind_max+1])
                phi_[i_ky] = phi_[i_ky] / renorm

        elif self.pnt.is3d:
            phi_ = np.zeros(len(range(self.kyind_min, self.kyind_max+1)), dtype=np.complex64)
            phi_c = np.fft.fft(phi, axis=1)/self.pnt.ny0
            jaco_xz = np.mean(geometry.jacobian.T, axis=1)
                        # loop over all ky modes
            for i_ky, j in enumerate(range(self.kyind_min, self.kyind_max+1)):
                for i in range(0, self.pnt.nx0-1):
                    renorm = renorm + np.dot(np.squeeze(phi_c[i, i_ky, self.zind_min:self.zind_max+1]**(2 * self.weight)), jaco_xz[i, self.zind_min:self.zind_max+1])
                    phi_[i_ky] = phi_[i_ky] + \
                        np.dot(np.squeeze(phi_c[i, i_ky, self.zind_min:self.zind_max+1]**(2 * self.weight) * phi_c[i, i_ky, self.zind_min:self.zind_max+1]), jaco_xz[i, self.zind_min:self.zind_max+1])
                phi_[i_ky] = phi_[i_ky] / renorm
        else:
            raise NotImplementedError("Frequency calculation not implemented")

        self.phi.append(phi_)
        
    def dict_to_mgkdb(self):
        Diag_dict = {}
        Diag_dict['opts']=self.opts
        Diag_dict['ky'] = self.ky
        Diag_dict['kyind_min'] = self.kyind_min
        Diag_dict['kyind_max'] = self.kyind_max
        Diag_dict['phi'] = self.phi
        
        return Diag_dict
        
        

    def plot(self, time_requested, output=None, out_folder=None, terminal=None, suffix = None):

        Freq_arr = []
        Grate_arr = []
        ky_arr = []

        def extract_freq(phi_pt, times_req, act_ky, plot_all, output=None):
            #   Amplitude and phase
            freq_trace = np.imag(np.diff(np.log(phi_pt)) / np.diff(times_req))
            gamma_trace = np.real(np.diff(np.log(phi_pt)) / np.diff(times_req))

            ph_f = np.angle(phi_pt)
            ampl_f = np.abs(phi_pt)

            #   linear regression of phase
            M = np.column_stack((np.ones(times_req.size), times_req))

            coeffsFreq = linalg.lstsq(M, ph_f)[0]
            coeffsAmp = linalg.lstsq(M, np.log(ampl_f))[0]

            Freq = coeffsFreq[1]
            Grate = coeffsAmp[1]

            Freq_arr.append(Freq)
            Grate_arr.append(Grate)

            if output:
                str_out = "ky mode {}: (Frequency, GrowthRate) {} {}\n".format(act_ky, Freq, Grate)
                output.info_txt.insert(END, str_out)

            if plot_all:
                ttl = r"$k_y\rho={:6.3f} $".format(act_ky)

                fig = plt.figure()

                ax_1 = fig.add_subplot(3, 1, 1)
                ax_1.plot(times_req, ampl_f, ls="-", color='b')
                ax_1.plot(times_req, np.exp(coeffsAmp[0] +
                                            coeffsAmp[1] * times_req),
                          ls="-", color='r')
                ax_1.set_xlabel(r"$t  c_{ref} / L_{ref}$")
                ax_1.set_yscale("log")
                ax_1.set_ylabel(r'$|\phi|$')
                ax_1.set_title(ttl)
                ax_1.text(0.05, 0.8, r'$\gamma={:4.3f}'.format(Grate) +
                          ' c_{ref}/L_{ref}$', transform=ax_1.transAxes)

                ax_2 = fig.add_subplot(3, 1, 2)
                ax_2.plot(times_req, ph_f, ls="-", color='b')
                ax_2.plot(times_req, coeffsFreq[0] + coeffsFreq[1] * times_req,
                          ls="--", color='r')
                ax_2.set_xlabel(r"$t  c_{ref} / L_{ref}$")
                ax_2.set_ylabel(r'$arg(\phi)$')
                ax_2.text(0.05, 0.8, r'$\omega={:4.3f}'.format(Freq) +
                          ' c_{ref}/L_{ref}$', transform=ax_2.transAxes)

                ax_3 = fig.add_subplot(3, 1, 3)
                ax_3.plot(times_req, np.abs(np.real(phi_pt)), ls='-',
                          color='b', label=r"$Re(|\phi|)$")
                ax_3.plot(times_req, np.abs(np.imag(phi_pt)), ls='--',
                          color='r', label=r"$Imag(|\phi|)$")
                ax_3.set_xlabel(r"$t  c_{ref} / L_{ref}$")
                ax_3.set_yscale("log")
                ax_3.legend()
                
                if out_folder is not None:
                    fig.savefig(os.path.join(out_folder, 'FreqGrowthRate_phi{}.png'.format(suffix)), bbox_inches='tight')                  
                    plt.close(fig)
                else:
                    fig.show()

                fig_t = plt.figure()
                a_1 = fig_t.add_subplot(2, 1, 1)
                a_1.plot(times_req[:-1], freq_trace, ls="-", color='b')
                a_1.set_xlabel(r"$t c_{ref}/L_{ref}$")
                a_1.set_ylabel(r"$\omega_r c_{ref}/L_{ref}$")
                a_1.set_title(ttl)

                a_2 = fig_t.add_subplot(2, 1, 2)
                a_2.plot(times_req[:-1], gamma_trace, ls="-", color='b')
                a_2.set_xlabel(r"$t c_{ref}/L_{ref}$")
                a_2.set_ylabel(r"$\gamma c_{ref}/L_{ref}$")
                
                if out_folder is not None:
                    fig_t.savefig(os.path.join(out_folder, 'FreqGrowthRate_OmegaGamma{}.png'.format(suffix)), bbox_inches='tight') 
                    plt.close(fig_t)
                else:
                    fig_t.show()

        self.plotbase = Plotting()
        plot_all = 1 if self.opts['show']['value'] else 0
        for ky in range(0, len(self.phi[0][:])):
            if (np.absolute(self.ky[ky]) <= np.absolute(self.ky[self.kyind_max])
                and np.absolute(self.ky[ky]) >= np.absolute(self.ky[self.kyind_min])):

                phi_trace = np.array([x[ky] for x in self.phi])
                act_ky = self.ky[ky]
                extract_freq(phi_trace, time_requested, act_ky, plot_all, output=None)
                ky_arr.append(act_ky)

        fig_f = plt.figure()
        a_1 = fig_f.add_subplot(2, 1, 1)
        a_1.plot(ky_arr, Grate_arr, 'bo')
        a_1.set_xlabel(r"$k_y$")
        a_1.set_ylabel(r"$\gamma c_{ref}/L_{ref}$")
        a_2 = fig_f.add_subplot(2, 1, 2)
        a_2.plot(ky_arr, Freq_arr, 'bo')
        a_2.set_xlabel(r"$k_y$")
        a_2.set_ylabel(r"$\omega c_{ref}/L_{ref}$")
        
        if out_folder is not None:
            fig_f.savefig( os.path.join(out_folder, 'FreqGrowthRate_OmegaGamma_arr{}.png'.format(suffix)), bbox_inches='tight') 
            plt.close(fig_f)
        else:   
            fig_f.show()

#        if self.opts['save_h5']['value']:
#            file = h5py.File(out_folder / 'frequency_growth_rate.h5', 'w')
#            file["/ky"] = ky_arr
#            file["/gamma"] = Grate_arr
#            file["/omega"] = Freq_arr
#            file.close()
            
    def save(time_requested, output=None, out_folder=None):
        pass

    def finalize():
        pass



""" Class to do plots of the anisotropy of B """
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from multiprocessing.dummy import Pool

from .baseplot import Plotting
from ..data.slices import MomFieldSlice

from .comm import DiagSpace
import .averages as averages
import .fourier as fourier
from ..utils import errors as err

class Anisotropy(object):
    """ Class for evaluating the scale-dependent anisotropy of magnetic field
    structures.
    B_L: local mean magnetic field
    b_l: local fluctuating field
    B_l: local field without filter

    :param commonlist: CommonData objects of the runs to plot
    :param rundatafiles: RunDataFiles objects of the runs to plot
    """

    def __init__(self, commonlist, rundatafiles, species=None):
        super().__init__()
        self.rundatafiles=rundatafiles
        self.cm = commonlist
        self.species = species

        #use fewer wavenumbers than ky modes for better statistics
        self.n_k=np.amin((int(2**np.rint(np.log2(self.cm.pnt.nky0/2))),32))
        print("Using {} k values for anisotropy diagnostic.".format(self.n_k))

        #calculate normalizing factor from an isotropic injection scale
        #numbers for solar wind according to Howes JGR 2008!
        k0=1e-4*np.sqrt(2) #isotropic injection scale at k*rho_i=10^-4 (rho_i defined with sqrt(2) by Howes)
        kymin=self.cm.pnt.kymin
        #assume Goldreich-Sridhar scaling of kpar~kperp^2/3 for MHD range (sda=scale-dependent anisotropy)
        #alternative: Boldyrev kpar~kperp^1/2
        model='GS'
        mhd_sda={'GS': 2./3, 'B': 1./2}
        names={'GS': 'Goldreich-Sridhar', 'B': 'Boldyrev'}
        kpar0=(kymin/k0)**(mhd_sda[model])*k0
        print("Calculating normalization factor according to {0} scaling. Result: epsilon = {1}".format(names[model],kpar0))
        #from GENE's box definitions, the normalizing epsilon factor equals kpar0
        self.epsl=kpar0

        self.nprocs=4
        self.k_array=np.zeros(self.n_k)
        self.kpar_av=np.zeros(self.n_k)
        #Cho & Lazarian: filter_factor=2
        self.filter_factor=2.

        #execute
        self.eval_kpar(species=species, rundatafiles=rundatafiles)


    def _fetch_moms(self, species, rundatafiles):
        """ Get mom and field slice object for species"""
        diagspace = DiagSpace(self.cm.spatialgrid, x_fourier=True, y_fourier=True, z_fourier=True)
        if self.cm.electromagnetic:
            moms={"apar": None}
            if self.cm.bpar:
                moms.update({"bpar": None})

        with Pool() as p:
            mom_temp = p.map(
                lambda mom: MomFieldSlice(self.cm, mom, species, diagspace, rundatafiles,
                                          modifier_func=lambda dum: dum), moms)
        for imom, mom in enumerate(moms):
            moms[mom] = mom_temp[imom]
        return moms

    def calc_individual_kpar(self,kval,a_par_fft,b_par_fft):
        k=self.k_array.tolist().index(kval)
        #print('k = {}'.format(kval))
        b_shape=np.shape(b_par_fft)
        #local mean field
        B_L=np.zeros((3,b_shape[0],b_shape[1],b_shape[2]),dtype=np.complex128)
        #local fluctuating field
        b_l=np.zeros((3,b_shape[0],b_shape[1],b_shape[2]),dtype=np.complex128)
        #aux array of ones with shape of b_l[0], used for array manipulations
        ones_b_l=np.ones(b_l[0].shape,dtype=np.float64)

        #3-d array of absolute value of k vector
        k_magn=np.sqrt(self.kx[:,None,None]**2+self.ky[None,:,None]**2+self.kz[None,None,:]**2*self.epsl**2)

        #Filters according to Cho & Lazarian 2004:
        #add to local mean field where k_magn <= kval/filter_factor
        ff=self.filter_factor
        B_L[0]=np.where(k_magn <= kval/ff, 1j*self.ky[None,:,None]*a_par_fft, 0.)
        B_L[1]=np.where(k_magn <= kval/ff, -1j*self.kx[:,None,None]*a_par_fft, 0.)
        B_L[2]=np.where(k_magn <= kval/ff, b_par_fft, 0.)

        #add to local fluctuating field where kval/2 < k_magn <= 2*kval
        b_l[0]=np.where(np.logical_and(k_magn>kval/ff,k_magn<=ff*kval), 1j*self.ky[None,:,None]*a_par_fft, 0.)
        b_l[1]=np.where(np.logical_and(k_magn>kval/ff,k_magn<=ff*kval), -1j*self.kx[:,None,None]*a_par_fft, 0.)
        b_l[2]=np.where(np.logical_and(k_magn>kval/ff,k_magn<=ff*kval), b_par_fft, 0.)

        #multiply by epsilon to transform to background units
        B_L*=self.epsl
        b_l*=self.epsl

        #compute nabla(b_l) matrix, z derivatives need extra epsilon factor to
        #compensate for different normalization
        #the grad_k step is a little awkward, but reduces the size of the code below
        grad_k=np.array([self.kx[:,None,None]*ones_b_l,
                         self.ky[None,:,None]*ones_b_l,
                         self.kz[None,None,:]*ones_b_l*self.epsl])
        grad_b_l_x=1j*grad_k*b_l[None,0,:,:,:]
        grad_b_l_y=1j*grad_k*b_l[None,1,:,:,:]
        grad_b_l_z=1j*grad_k*b_l[None,2,:,:,:]
        del ones_b_l, grad_k

        #Transform B_L, b_l, and grad(b_l) to real space for dot product
        B_L=fourier.kz_to_z(B_L,self.cm.pnt.nz0)
        b_l=fourier.kz_to_z(b_l,self.cm.pnt.nz0)
        grad_b_l_x=fourier.kz_to_z(grad_b_l_x,self.cm.pnt.nz0)
        grad_b_l_y=fourier.kz_to_z(grad_b_l_y,self.cm.pnt.nz0)
        grad_b_l_z=fourier.kz_to_z(grad_b_l_z,self.cm.pnt.nz0)

        B_L=np.fft.irfft2(B_L, axes=(-3,-2))*2*self.cm.pnt.nx0*self.cm.pnt.nky0
        b_l=np.fft.irfft2(b_l, axes=(-3,-2))*2*self.cm.pnt.nx0*self.cm.pnt.nky0
        grad_b_l_x=np.fft.irfft2(grad_b_l_x, axes=(-3,-2))*2*self.cm.pnt.nx0*self.cm.pnt.nky0
        grad_b_l_y=np.fft.irfft2(grad_b_l_y, axes=(-3,-2))*2*self.cm.pnt.nx0*self.cm.pnt.nky0
        grad_b_l_z=np.fft.irfft2(grad_b_l_z, axes=(-3,-2))*2*self.cm.pnt.nx0*self.cm.pnt.nky0

        #add background field
        B_L[2]+=1.0

        # Calculate kpar (Groselj)
        compx = np.sum(B_L*grad_b_l_x,axis=0)
        compy = np.sum(B_L*grad_b_l_y,axis=0)
        compz = np.sum(B_L*grad_b_l_z,axis=0)
        del grad_b_l_x, grad_b_l_y, grad_b_l_z

        dotproductabs = compx**2+compy**2+compz**2
        numerator=np.average(dotproductabs,axis=(-3,-2,-1))
        denom1=np.average(B_L[0]**2+B_L[1]**2+B_L[2]**2,axis=(-3,-2,-1))
        denom2=np.average(b_l[0]**2+b_l[1]**2+b_l[2]**2,axis=(-3,-2,-1))
        denominator=denom1 * denom2
        del B_L, compx, compy, compz, b_l
        return np.average(np.sqrt(numerator/denominator))

    def eval_kpar(self,species, rundatafiles):

        """ Plot contour of kpar-kperp-spectrum
        """
        #get kx,ky,kz via self.cm.spatialgrid
        self.kx=self.cm.spatialgrid.kx_fftorder
        self.ky=self.cm.spatialgrid.ky
        self.kz=self.cm.spatialgrid.kz

        #We assume an isotropic box and take ky_min and ky_max
        #as constraints for the perpendicular wavenumber range.
        #For the lowest ky modes, the background filter will be incomplete
        #and dominated by the actual background field.
        self.k_array=np.logspace(np.log10(self.ky[1]),np.log10(self.ky[-1]),self.n_k)
        ###################################
        moms = self._fetch_moms(species,rundatafiles)
        moms["apar"].check_times()
        pos = moms["apar"].calc_positions()
        self.timearray = np.take(moms["apar"].timearray, pos)

        self.kpar=np.zeros((len(self.timearray),len(self.k_array)),dtype=np.float64)
        for time in self.timearray[::max(int(len(self.timearray)/20),1)]:
            tind=self.timearray.tolist().index(time)
            print("Time: {}".format(time))

            # get b-field from field file
            b_par_fft=moms["bpar"].generate_slice_attime(time)
            a_par_fft=moms["apar"].generate_slice_attime(time)

            # Calculate Anisotropy with Fourier based filter method (Cho & Lazarian 2004)
            # Use limited number of threads, since this can easily use a whole node's memory for a large simulation
            with Pool(self.nprocs) as p:
                self.kpar[tind] = p.map(
                    lambda kval: self.calc_individual_kpar(kval, a_par_fft, b_par_fft), self.k_array)
        #compute time averaged kpar spectrum
        self.kpar_av = averages.mytrapz(self.kpar, self.timearray)*max(int(len(self.timearray)/20),1)


class PlotAnisotropy(Plotting):
    def __init__(self,commonlist,anisotropyserieslist):
        super().__init__()
        self.anisotropylist = anisotropyserieslist
        self.cm = commonlist
    def createfigs(self):
        for cm, anisotropyseries in zip(self.cm,self.anisotropylist):
            k_arr=anisotropyseries.k_array
            ff=anisotropyseries.filter_factor
            fig = plt.figure(figsize=(8.,5.))
            ax = fig.add_subplot(111)
            ax.loglog(k_arr, anisotropyseries.kpar_av,'o',label="$k_\parallel$")
            #find index closest to kperp=1
            ind=np.argmin(np.abs(k_arr-1))
            #get kpar value at kperp=1 to plot fits close to that curve
            kp_1=anisotropyseries.kpar_av[ind]
#            ax.loglog(k_arr, k_arr**(5./3)*0.1,'-',label="$k_\perp^{5/3}$")
#            ax.loglog(k_arr, k_arr**(3./3)*0.1,'-',label="$k_\perp^{3/3}$")
            ax.loglog(k_arr, k_arr**(2./3)*kp_1*1.2,'-',label="$k_\perp^{2/3}$")
#            ax.loglog(k_arr, k_arr**(1./2)*kp_1*0.8,'-',label="$k_\perp^{1/2}$")
            ax.loglog(k_arr, k_arr**(1./3)*kp_1*1.2,'--',label="$k_\perp^{1/3}$")
            xlim=ax.get_xlim()
            ylim=ax.get_ylim()
            #draw shaded rectangles in wavenumber ranges where filter ranges are not completely filled with modes
            rect_l=mpatches.Rectangle((xlim[0],ylim[0]),ff*np.amin(k_arr)-xlim[0],ylim[1]-ylim[0],fc=(0.5,0.5,0.5,0.25))
            rect_r=mpatches.Rectangle((np.amax(k_arr)/ff,ylim[0]),xlim[1]-np.amax(k_arr)/ff,ylim[1]-ylim[0],fc=(0.5,0.5,0.5,0.25))
            ax.add_patch(rect_l)
            ax.add_patch(rect_r)
            ax.set_xlabel(r"$k_{\perp}\rho_i$")
            ax.set_ylabel(r"$k_{\parallel}\rho_i$")
            ax.set_title(r"Filter Method", y=1.03)
            ax.legend(loc='upper left', fontsize="medium")
            fig.tight_layout()
            fig.savefig('filter_method{}.pdf'.format(cm.fileextension))

            # name output file
#            file=open('data_kpar.txt','w')
#            file.write('#%16s %16s\n'%('kperp','kpar'))
#            for i in range(len(k_arr)):
#                file.write('%16.8e %16.8e\n'%(k_arr[i], anisotropyseries.kpar_av[i]))
#            file.close()

#    @classmethod
#    def show(cls):
#        plt.show()



""" Module containing the contour diagnostics"""
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import numpy as np
import putils.averages as averages

from .diagnostic import Diagnostic
from .baseplot import Plotting
from .diagspace import DiagSpace


class DiagContours(Diagnostic):
    """ Diagnostic for contours in the xy and xz plane"""

    def __init__(self, avail_vars=None, specnames=None):
        super().__init__()
        self.name = 'Contours'
        self.tabs = ['fluxtube', 'xglobal']

        self.help_txt = """Plot contours on xy and xz planes
                        \n
                        \nQuantity : quantity to plot
                        \nx ind : index of radial position (def. all)
                        \ny ind : index of binormal position (def. 0)
                        \nz ind : index of parallel position (def. outboard midplane)
                        \nspec: which species to plot (def. all if spec dependent)
                        \nFourier x : spectral in x direction (def. False)
                        \nFourier y : spectral in y direction (def. False)
                        \nSave h5 : save hdf5 file (def. False)
                        \nt avg : time averaged (def. False)"""

        self.avail_vars = avail_vars
        self.specnames = specnames

        self.opts = {"quant": {'tag': "Quant", 'values': None, 'files': ['mom', 'field']},
                     "spec": {'tag': "Spec", 'values': self.list_specnames()},
                     "xind": {'tag': "x ind", 'values': None},
                     "yind": {'tag': "y ind", 'values': None},
                     "zind": {'tag': "z ind", 'values': None},
                     "f_x": {'tag': "Fourier x", 'values': [False, True]},
                     "f_y": {'tag': "Fourier y", 'values': [False, True]},
                     'save_h5': {'tag': "Save h5", 'values': [False, True]},
                     't_avg': {'tag': "t avg", 'values': [False, True]}, }

        self.set_defaults()

        self.options_gui = Diagnostic.OptionsGUI()
        self.contours = {}
        self.contours_parallel = {}

    def set_options(self, run_data, specnames=None):

        if self.opts['quant']['value'] is None:
            raise RuntimeError("No quantities given for contours")

        # parallel position, unset mean outboard midplane, -1 all
        self.z_ind = self.opts['zind']['value']
        self.z_avg = False
        if self.zind == -1:
            self.z_avg = True

        # radial position, unset or -1 means all
        self.x_ind = self.opts['xind']['value']
        self.x_avg = False
        if self.x_ind == -1:
            self.x_avg = True

        # radial position, unset or -1 means all
        self.y_ind = self.opts['yind']['value']
        self.y_avg = False
        if self.y_ind == -1:
            self.y_avg = True

        # Fourier in x
        self.x_fourier = self.opts['f_x']['value']
        # Fourier in y
        self.y_fourier = self.opts['f_y']['value']

        self.run_data = run_data

        # This is special since we dont save stuff by default, so just do what we need
        self.get_spec_from_opts()

        # toggle the reader
        self.get_needed_vars()

        # cast the output as a dictionary of lists to ease the rest
        #        TODO there is probably a best structure for this....
        self.contours = {}
        for file in self.needed_vars:
            # loop over quaitites in that file
            for quant in self.needed_vars[file]:
                if self.needed_vars[file][quant]:
                    if file == 'field':
                        self.contours[quant] = []
                    else:
                        for spec in self.specnames:
                            self.contours[quant + '#' + spec] = []

        return self.needed_vars

    def execute(self, data, parameters, geometry, spatialgrid,  step):
        """ not much to do other than appending data """

        # loop over files
        for file in self.needed_vars:
            # loop over quaitites in that file
            for quant in self.needed_vars[file]:
                # loop over quaitites in that file
                if self.needed_vars[file][quant]:
                    if file == 'field':
                        # no species dependency
                        data_in = getattr(data.field,'quant')(step.time, step.field, parameters)
                        data_in = self._apply_fouriertransforms(data_in, geometry)
                        self.contours[quant].append(data_in)
                    else:
                        # spec dependent
                        for spec in self.specnames:
                            data_in = getattr(getattr(data, file + '_' + spec), quant)(
                                    step.time, step.field, parameters)

                            data_in = self._apply_fouriertransforms(data_in, geometry)
                            self.contours[quant + '#' + spec].append(data_in)

    def plot(self, time_requested, output=None, out_folder=None):
        """ Not much to explain """

        plotbase = Plotting()
        if self.run_data.is3d:
            plotbase.titles.update(
                    {"phi": r"$\phi$", "n": r"$n$", "u_par": r"u_{//}", "T_par": r"T_{//}",
                     "T_perp": r"T_{\perp}", "Q_es": r"$Q_{es}$", "Q_em": r"$Q_{em}$",
                     "Gamma_es": r"$\Gamma_{es}$", "Gamma_em": r"$\Gamma_{em}$"})
        else:
            plotbase.titles.update(
                    {"phi": r"$\phi$", "A_par": r"$A_{//}$", "B_par": r"$B_{//}$", "dens": r"$n$",
                     "T_par": r"T_{//}", "T_perp": r"T_{\perp}", "u_par": r"u_{//}",
                     "q_par": r"q_{//}", "q_perp": r"q_{\perp}"})

        z_lbl = r"$z$"
        z = self.run_data.spatialgrid.z

        if self.opts['f_x']['value']:
            x_lbl = r"$k_x\rho_{ref}$"
            x = self.run_data.spatialgrid.kx_fftorder
        else:
            x_lbl = r"$x/\rho_{ref}$"
            x = self.run_data.spatialgrid.x

        if self.opts['f_y']['value']:
            y_lbl = r"$k_y\rho_{ref}$"
            y = self.run_data.spatialgrid.ky
        else:
            y_lbl = r"$y/\rho_{ref}$"
            y = self.run_data.spatialgrid.y

        is_f = self.opts['f_x']['value'] or self.opts['f_y']['value']

        if self.opts['t_avg']['value']:
            # loop over file
            for quant in self.contours:
                ind_str = quant.find('#')
                if ind_str == -1:
                    ttl = plotbase.titles[quant]
                else:
                    ttl = plotbase.titles[quant[0:ind_str]] + " " + quant[ind_str + 1:]

                fig = plt.figure()
                ax = fig.add_subplot(111)
                plotbase.plot_contour_quant(ax, fig, x, y,
                                            averages.mytrapz(self.contours[quant], time_requested),
                                            is_f, x_lbl, y_lbl, ttl, is_bidirect=True)

        else:
            for quant in self.contours:
                ind_str = quant.find('#')
                if ind_str == -1:
                    ttl = plotbase.titles[quant]
                else:
                    ttl = plotbase.titles[quant[0:ind_str]] + " " + quant[ind_str + 1:]

                fig = plt.figure()
                ax = fig.add_subplot(111)
                plotbase.plot_contour_quant_timeseries(ax, fig, x, y, self.contours[quant], is_f,
                                                       time_requested, x_lbl, y_lbl, ttl, is_bidirect=True)

                fig = plt.figure()
                ax = fig.add_subplot(111)
                print(np.size(self.contours_parallel[quant]))
                plotbase.plot_contour_quant_timeseries(ax, fig, x, z, self.contours_parallel[quant], is_f,
                                                       time_requested, x_lbl, z_lbl, ttl, is_bidirect=True)

        plt.show()

    def _apply_foutier_Transforms(self, data, geometry):
        



# -*- coding: utf-8 -*-

from tkinter import END
import matplotlib.pyplot as plt
import numpy as np
from .baseplot import Plotting
from .diagnostic import Diagnostic


class DiagFieldMomSnapshots(Diagnostic):

    def __init__(self):
        super().__init__()
        print("Initializing Field Mom Snapshots")
        self.name = 'Field Moment Snapshots'
        self.tabs = ['fluxtube']

    def get_info(self):
        self.need_file={'field': True,
                        'mom': True}

        return self.need_file
    
    def setup_options(self, run):
        self.specnames = run.parameters.pnt.specnames
        self.time = []
        self.n_fields = run.parameters.pnt.n_fields
        self.fm_snapshots = []
        self.fm_final = {}
        self.n_moms = run.parameters.pnt.n_moms
        self.n_spectra = int(run.parameters.pnt.n_fields + run.parameters.pnt.n_moms*len(self.specnames))
        self.kx=run.spatialgrid.kx
        self.ky=run.spatialgrid.ky
        self.zgrid = run.spatialgrid.z
        self.momlist = ["dens", "tpar", "tperp", "qpar", "qperp", "upar", 'densI1', 'TparI1', 'TppI1']
        self.field_mom_names=[]
        self.field_mom_names.append('phi')
        if run.parameters.pnt.nonlinear == 'T' or run.parameters.pnt.nonlinear == True:
            self.nonlinear = True
        else:
            self.nonlinear = False
        if self.n_fields > 1:
            self.field_mom_names.append('apar')
        if self.n_fields > 2:
            self.field_mom_names.append('bpar')
        for i in self.specnames:
            for j in range(self.n_moms):
                self.field_mom_names.append(self.momlist[j]+i)
        
        
    def execute(self, data, run, steps, extensions,time_point):
        #print('fms 47')
        self.time.append(time_point)
        nz0 = run.parameters.pnt.nz0
        #print(run.parameters.pnt.nonlinear)
        #print('nonlinear',nonlinear)
        #dummy = input("press key")
        self.zgrid_sparse = [self.zgrid[0],self.zgrid[int(nz0/4)],self.zgrid[int(nz0/2)],self.zgrid[int(3*nz0/4)]]
        if self.nonlinear:
            fm_sparse = np.empty((self.n_spectra,run.parameters.pnt.nx0,run.parameters.pnt.nky0,4),dtype='complex')
   
        #print('fms 54')
        #print('shape of fm_sparse',np.shape(fm_sparse))
        tmp=data.field.phi(step=steps['field'], extension=extensions['field'])
        #print('shape of tmp',np.shape(tmp))
        self.fm_final[self.field_mom_names[0]] = tmp
        if self.nonlinear:
            fm_sparse[0,:,:,0] = tmp[:,:,0]
            fm_sparse[0,:,:,1] = tmp[:,:,int(nz0/4)]
            fm_sparse[0,:,:,2] = tmp[:,:,int(nz0/2)]
            fm_sparse[0,:,:,3] = tmp[:,:,int(3*nz0/4)]
        #print('fms 60')
        if self.n_fields>1:
            tmp=data.field.A_par(step=steps['field'], extension=extensions['field'])
            if self.nonlinear:
                fm_sparse[1,:,:,0] = tmp[:,:,0]
                fm_sparse[1,:,:,1] = tmp[:,:,int(nz0/4)]
                fm_sparse[1,:,:,2] = tmp[:,:,int(nz0/2)]
                fm_sparse[1,:,:,3] = tmp[:,:,int(3*nz0/4)]
            self.fm_final[self.field_mom_names[1]] = tmp
        if self.n_fields>2:
            tmp=data.field.B_par(step=steps['field'], extension=extensions['field'])
            if self.nonlinear:
                fm_sparse[2,:,:,0] = tmp[:,:,0]
                fm_sparse[2,:,:,1] = tmp[:,:,int(nz0/4)]
                fm_sparse[2,:,:,2] = tmp[:,:,int(nz0/2)]
                fm_sparse[2,:,:,3] = tmp[:,:,int(3*nz0/4)]
            self.fm_final[self.field_mom_names[2]] = tmp
             
        #print('fms 73')
        for i_spec, spec in enumerate(run.parameters.specnames):
            for i_quant, quant in enumerate(data.mom[i_spec].varnames.values()):
                tmp = getattr(data.mom[i_spec],quant)(step=steps['mom'], extension=extensions['mom'])
                if self.nonlinear:
                    fm_sparse[self.n_fields+len(self.specnames)*i_spec+i_quant,:,:,0] = tmp[:,:,0]
                    fm_sparse[self.n_fields+len(self.specnames)*i_spec+i_quant,:,:,1] = tmp[:,:,int(nz0/4)]
                    fm_sparse[self.n_fields+len(self.specnames)*i_spec+i_quant,:,:,2] = tmp[:,:,int(nz0/2)]
                    fm_sparse[self.n_fields+len(self.specnames)*i_spec+i_quant,:,:,3] = tmp[:,:,int(3*nz0/4)]
                self.fm_final[self.field_mom_names[self.n_fields+len(self.specnames)*i_spec+i_quant]] = tmp
                        
        #print('fms 82')
        if self.nonlinear:
            self.fm_snapshots.append(fm_sparse)
                
    def dict_to_mgkdb(self):
        
        #print('fms 83')
        fm_out = {}
        if self.nonlinear:
            fm_out['Description'] = 'All fields and moments at selected time points at z = -pi,-pi/2,0,pi/2.  The structure is [field_or_moment][time list][kx,ky,four z points]. The time points are in [time].  The names of the fields and moments are in [field_mom_names].  The appropriate grids are in [kxgrid], [kygrid], and [zgrid_sparse].  The full 3D data for each field and moment at the final time point is stored in [field_mom_final][name]'
        else:
            fm_out['Description'] = 'For linear: The final time point is in [time].  The names of the fields and moments are in [field_mom_names].  The appropriate grids are in [kxgrid], [kygrid], and [zgrid].  The full 3D data for each field and moment at the final time point is stored in [field_mom_final][name]'
        fm_out['time'] = self.time
        fm_out['field_mom_names'] = self.field_mom_names
        fm_out['kxgrid'] = np.roll(self.kx,int(len(self.kx)/2)+1)
        fm_out['kygrid'] = self.ky
        if self.nonlinear:
            fm_out['zgrid_sparse'] = self.zgrid_sparse
        fm_out['zgrid'] = self.zgrid
        fm_out['field_mom_final'] = self.fm_final
        #print('fms 91')
        if self.nonlinear:
            for i in range(len(self.field_mom_names)):
                fm_out[self.field_mom_names[i]] = []
                for j in range(len(self.fm_snapshots)):
                    fm_out[self.field_mom_names[i]].append(self.fm_snapshots[j][i,:,:,:])
        return fm_out
        
    def plot(self, time_requested, output=None, out_folder=None):
        """ For each selected species we have one figure with six subplots.
            Left is vs. kx, right vs. ky; columnwise we plot, log-log, log-lin, lin-in
            Dashed lines are negative values in log-log plot
            Dashed lines are k multiplied values in log-lin plot"""
        pass





# -*- coding: utf-8 -*-

from bisect import bisect_left, bisect_right
from copy import deepcopy
from dataclasses import dataclass

import numpy as np


class Loader:
    """ This does the nasty job. We need for a given timestep to be able to load
        any kind of data. The idea is the following:
            1- the user sets folder and runs
            2- GUI determines which files and variables are available
            3- GUI fetches all times from all the files
            4- user selects a time window and a series of diagnostics
            5- GUI takes the subset of files that are needed
            5- GUI takes the requested subinterval in time
            6- loop over all times

            -- I win.
            """

    def __init__(self, diags, run, data, t_start=None, t_end=None):
        self.t_start = t_start
        self.t_end = t_end
        # we need to call each diagnostic, get the file it needs
        self.specnames = run.specnames
        self.needed_vars = []
        for diag in diags:
            self.needed_vars.append(diag.set_options(run, None))

        # we need to decide which files need to be loaded by joining needed_vars
        #    We simply join the whole but keep the original list
        global_vars = deepcopy(self.needed_vars[0])
        for file_key in global_vars.keys():  # $lis(self.needed_vars):
            for quant_key in global_vars[file_key].keys():
                for i_diag_needs in list(self.needed_vars):
                    global_vars[file_key][quant_key] = global_vars[file_key][quant_key] or \
                                                        i_diag_needs[file_key][quant_key]

        # now we set needed_field to true or false depending if any diag needs it
        self.needed_files = {}
        #        TODO what if data is not available
        # loop over available fields
        for k in data.avail_vars.keys():
            # loop over requested fields
            if k in global_vars.keys():
                if any(global_vars[k].values()):
                    self.needed_files[k] = True
                else:
                    self.needed_files[k] = False
            else:
                self.needed_files[k] = False

    def set_interval(self, data, t_start, t_end, step):

        class ProcessableTime:
            def __init__(self, time):
                self.time = time

        # Here we intersect all times with the one we want
        all_times = np.array([])

        # loop over files that are necessary
        for file in self.needed_files:

            if self.needed_files[file]:
                if len(all_times) == 0:
                    all_times = getattr(getattr(data.avail_times, file), 'times')
                else:
                    all_times = np.intersect1d(all_times,
                                               getattr(getattr(data.avail_times, file), 'times'))

        # now reduce the interval
        i_s = bisect_left(all_times, t_start)
        i_e = bisect_right(all_times, t_end)
        if i_s == len(all_times):
            i_s -= 1
        all_times = np.array(all_times[i_s:i_e:step])
        self.t_start = all_times[0]
        self.t_end = all_times[-1]

        # and now fill the step list
        self.processable_times = all_times
        self.processable = [ProcessableTime(t) for t in all_times]

        for file in self.needed_files:
            if self.needed_files[file]:
                tmp_t = getattr(getattr(data.avail_times, file), 'times')
                ind = np.arange(tmp_t.shape[0])[np.in1d(tmp_t, all_times)]
                my_steps = getattr(getattr(data.avail_times, file), 'steps')[ind]
                my_exts = [getattr(getattr(data.avail_times, file), 'files')[i] for i in ind]
                [setattr(self.processable[x], file, self.TimeStep(my_steps[x], my_exts[x][:]))
                 for x in range(all_times.size)]

        # NOTE this does a sync between all diags. meaning if there is diag1 that wants field
        # and diag2 that wants field and mom, diag1 will be executed only on the times of diag2
        # This is a loss of data. To fix this we need to keep track of whether a given diag wants
        # to use multiple files at once or not.

    @dataclass
    class TimeStep:
        step: int
        file: str



"""
Module handling neoclass files

"""
import csv
import numpy as np

from .base_file import TimeSeries


class NeoclassFile(TimeSeries):
    """NeoclassFile: Class to read a neoclass file from GENE

    :param filename: Typically neoclass.dat
    :param common: CommonData object of the run
    """
    def __init__(self, filename, common):
        super().__init__(filename, common)
        self.dataarray = []
        self.n_col = 4  # Particle, heat, momentum flux and bootstrap current
        self.n_spec = common.pnt.n_spec

    def generate_timeseries(self):
        """ Fill the NeoclassFile object with data """
        n_col = self.n_col
        try:
            with open(self.objectname) as ncfile:
                csvnrg = csv.reader(ncfile, delimiter=' ', skipinitialspace=True)
                for line in csvnrg:
                    if len(line) == 0:
                        continue
                    if len(line) == 1:
                        self.timearray.append(float(line[0]))
                        self.dataarray.append([[] for _ in range(self.n_spec)])
                        ispec = 0  # Reset index for the species at the current time step
                    elif len(line) == n_col:
                        self.dataarray[-1][ispec] = line
                        ispec += 1
                    else:
                        raise IOError("Incorrect number of columns")
        except IOError:
            raise IOError("neoclass file does not exist or has"
                          " wrong number of columns: {}".format(self.objectname))
        self.check_times()
        pos = self.calc_positions()
        self.dataarray = np.array(self.dataarray).astype(float, copy=False)
        # Reduce the nrgcols to only the required time frame
        self.dataarray = self.dataarray[pos, ...]
        self.timearray = np.array(self.timearray)[pos]



""" Module with the container objects for single simulation data"""
# -*- coding: utf-8 -*-
from pathlib import Path
from .par_io import Parameters
from .vsp_grid import VspGrid
from .spatial_grid import SpatialGrid
from ..data.profiledata import ProfileData
from .geom import Geometry
from .units import Units
import numpy as np
import os


class Run:
    ### Class containing basic information on a simulation run parameters. ###
    ##  in_folder [str]: directory containing gene output files
    ##  extension [str]: extension of run being processed, ex) 0001 or .dat
    ##
    ###################################################################

    def __init__(self, in_folder=None, extension='.dat'):
        if not os.path.isdir(in_folder):
            raise Exception(os.path.abspath(in_folder) + " does not exist or is undefined")
        
        self.in_folder = Path(in_folder)
        self.extension = extension
        self.parameters = Parameters(self.in_folder,self.extension)
        self.geometry = Geometry(self.in_folder,self.extension,self.parameters)
        self.spatialgrid = SpatialGrid(self.parameters)
        self.vspgrid = VspGrid(self.parameters)
        self.units = Units(self.parameters)
        self.profiles = ProfileData(self.in_folder,self.extension,self.parameters)
        omega_path = str(self.in_folder)+'/omega'+self.extension
        if os.path.isfile(omega_path):
            a = np.genfromtxt(omega_path)
            self.gamma = a[1]
            self.omega = a[2]
        else:
            self.gamma = False
            self.omega = False



# -*- coding: utf-8 -*-


class DiagSpace:
    """ Class defining the space that should be used in a diagnostic

    i.e. which intervals in x, y, z, Fourier or configuration space in x,y
    Velocity space is not handled by this class
    The ranges are given as tuples of the starting and stopping index (which is not included)
    Negative indices can be used in the usual Python way (-1 indexing the last element).
    Be aware that while [0:] includes the last element [0:-1] does not!
    Refer to https://docs.scipy.org/doc/numpy/reference/arrays.indexing.html for the details
    how indexing numpy arrays works.
    A binary flag indicates if an average is desired in that direction.
    """

    def __init__(self, rungrid, x_fourier, y_fourier, z_fourier=False, xrange=(None, None),
                 yrange=(None, None), zrange=(None, None), xavg=False, yavg=False, zavg=False):
        """ Standard constructor for DiagSpace

        :param rungrid: SpatialGrid object, supplies the numerical grid of the GENE run
        :param x_fourier: If true, operate in kx space, Fourier transform data if necessary
        :param y_fourier: If true, operate in ky space, Fourier transform data if necessary
        :param z_fourier: If true, operate in kz space, Fourier transform data if necessary
        :param xrange: x or kx index range as a tuple (i_min, i_max). Note that kx refers to the
        positive kx modes here.
        :param yrange: y or ky index range as a tuple (j_min, j_max)
        :param zrange: z index range as a tuple (k_min, k_max)

        Note that None is also a possible choice for the min and max values due to [:]
        giving the entire array.
        """
        self.xavg = xavg
        self.yavg = yavg
        self.zavg = zavg
        self.x_fourier = x_fourier
        self.y_fourier = y_fourier
        self.z_fourier = z_fourier
        if x_fourier:
            self.xslice = self._create_slice(xrange, rungrid.kx_pos)
        else:
            self.xslice = self._create_slice(xrange, rungrid.x)
        if y_fourier:
            self.yslice = self._create_slice(yrange, rungrid.ky)
        else:
            self.yslice = self._create_slice(yrange, rungrid.y)
        if z_fourier:
            self.zslice = self._create_slice(zrange, rungrid.kz)
        else:
            self.zslice = self._create_slice(zrange, rungrid.z)
        self.xyslice = (self.xslice, self.yslice)
        self.xzslice = (self.xslice, self.zslice)
        self.yzslice = (self.yslice, self.zslice)
        if not (xavg or yavg or zavg):
            self.diagslice = (self.xslice, self.yslice, self.zslice)
        elif xavg and not yavg and not zavg:
            self.diagslice = self.yzslice
        elif xavg and yavg and not zavg:
            self.diagslice = self.zslice
        elif xavg and not yavg and zavg:
            self.diagslice = self.yslice
        elif not xavg and not yavg and zavg:
            self.diagslice = self.xyslice
        elif not xavg and yavg and not zavg:
            self.diagslice = self.xzslice
        elif not xavg and yavg and zavg:
            self.diagslice = self.xslice
        else:
            self.diagslice = None
        self.rungrid = rungrid

    @staticmethod
    def _create_slice(coordrange, grid_direction):
        if grid_direction[coordrange[0]:coordrange[1]].size == 0:
            raise IndexError(
                    "The given range {} is empty or outside of the grid ({}:{}) of the run".format(
                            coordrange, 0, grid_direction.size - 1))
        return slice(coordrange[0], coordrange[1])



import numpy as np
import warnings
from ..utils.averages import av3d_by_switch
from ..utils.geom import Geometry
from .base_file import TimeSeries
from ..utils.fourier import kx_to_x, x_to_kx, ky_to_y, y_to_ky, z_to_kz


class MomFieldSlice(TimeSeries):
    """ Class to generate spatial slices of Mom or Field data"""

    def __init__(self, common, quantity, species, diagspace, rundatafiles,
                 modifier_func=lambda var: np.real(var*np.conj(var))):
        """ Constructor for MomFieldSlice

        :param common: CommonData object of the run
        :param quantity: Name of the variable to process
        :param species: Name of the species to process
        :param diagspace: Fourier or configuration space slice desired for x and y, are
        transforms necessary from the native format of the run?
        :param rundatafiles: A RunDataFiles object that provides the mom and field files necessary
        :param modifier_func: A function reference which should be applied to the 3d data first
        By default the squared absolute is calculated
        """
        super().__init__("slice", common)
        self.cm = common
        self.geom = Geometry(self.cm)
        self.diagspace = diagspace
        self.quantity = quantity
        if quantity in ["phi", "apar", "bpar"]:
            self.fileobject = rundatafiles.get_fileobject("field")
        elif quantity in ["dens", "tpar", "tperp", "qpar", "qperp", "upar",
                          "n00", "n20", "n02"]:
            self.fileobject = rundatafiles.get_fileobject("mom_{}".format(species))
        else:
            raise RuntimeError("Invalid variable name")
        self.modifier_func = modifier_func

    def generate_timeseries(self, sparsefactor=1):
        self.check_times()
        pos = self.calc_positions()
        self.dataarray = []
        new_timearray = []
        if sparsefactor != 1:
            pos = pos[::sparsefactor]
        for tind in pos:
            self.dataarray.append(self.generate_slice_attime(self.timearray[tind]))
            new_timearray.append(self.timearray[tind])
        self.timearray = np.array(new_timearray)
        self.dataarray = np.atleast_2d((np.array(self.dataarray)))

    def generate_slice_attime(self, time):
        """ Create a slice of the data at a fixed time with the desired dimensions

        :param time: The exact time to use
        """
        self.fileobject.set_time(time)
        var = self.fileobject.get_var(varname=self.quantity)
        averagedvar = av3d_by_switch(self.diagspace.xavg, self.diagspace.yavg,
                                         self.diagspace.zavg)(var, self.geom)
        averagedvar = self._apply_fouriertransforms(averagedvar)
        averagedvar = self.modifier_func(averagedvar)
        return averagedvar[self.diagspace.diagslice]

    def _apply_fouriertransforms(self, var):
        if self.diagspace.xavg and self.diagspace.yavg:  # Nothing to do if both are averaged
            return var
        xaxis = -3  # By default the last three dimensions of var are x, y, z.
        yaxis = -2  # This is changed if averages are applied
        zaxis = -1
        if self.diagspace.yavg:
            xaxis += 1
            yaxis += 1
        if self.diagspace.zavg:
            xaxis += 1
            yaxis += 1

        if self.diagspace.x_fourier != self.cm.x_local and not self.diagspace.xavg:
            if self.cm.x_local:
                var = kx_to_x(var, self.cm.pnt.nx0, axis=xaxis)
            else:
                var = x_to_kx(var, self.cm.pnt.nx0, axis=xaxis)
        if self.diagspace.y_fourier != self.cm.y_local and not self.diagspace.yavg:
            if self.cm.y_local:
                var = ky_to_y(var, self.cm.pnt.nky0, axis=yaxis)
            else:
                warnings.warn("y-global is not well tested", RuntimeWarning)
                var = y_to_ky(var, self.cm.pnt.nky0, axis=yaxis)
        if self.diagspace.z_fourier:
            var=z_to_kz(var, self.cm.pnt.nz0, axis=zaxis)
        return var

    def write_to_numpy(self, filename):
        self.generate_timeaverage()
        desc = """Slice diagnostic for {} (with {} applied), t={}, xavg,yavg,zavg={},{},{},"""
        """Range (x,y,z)={},{},{}""".format(self.quantity, self.modifier_func, self.get_minmaxtime(),
                                            self.diagspace.xavg, self.diagspace.yavg,
                                            self.diagspace.zavg, self.diagspace.xslice,
                                            self.diagspace.yslice, self.diagspace.zslice)
        np.savez_compressed(filename, description=desc, timearray=self.timearray,
                            dataarray=self.dataarray, timeaverage=self.timeaverage)

    def write_to_ascii(self, filename, t_average=False):
        if np.squeeze(self.dataarray[0, ...]).ndim > 2:
            raise RuntimeError("Unable to write 4d data (time + x,y,z) as ASCII file")
        if t_average:
            self.generate_timeaverage()
            np.savetxt(filename, np.squeeze(self.timeaverage),
                       header="Slice diagnostic for {} (with {} applied), averaged t={}\n".format(
                               self.quantity, self.modifier_func, self.get_minmaxtime()))
        else:
            with open(filename, "wb") as file:
                file.write(bytearray("""# Slice diagnostic for {} (with {} applied), t={},\n"""
                                     """# xavg,yavg,zavg={},{},{}, Range (x,y,z)={},{},
                                     {}\n""".format(
                        self.quantity, self.modifier_func, self.get_minmaxtime(),
                        self.diagspace.xavg, self.diagspace.yavg, self.diagspace.zavg,
                        self.diagspace.xslice, self.diagspace.yslice, self.diagspace.zslice),
                        encoding="utf-8"))
                for tind, time in enumerate(self.timearray):
                    np.savetxt(file, np.atleast_1d(np.squeeze(self.dataarray[tind])), fmt='%.9e',
                               header="{}".format(self.timearray[tind]))
                    file.write(b"\n")
